OUTPUT_A = 'outA'
OUTPUT_B = 'outB'
OUTPUT_C = 'outC'
OUTPUT_D = 'outD'

INPUT_1 = 'in1'
INPUT_2 = 'in2'
INPUT_3 = 'in3'
INPUT_4 = 'in4'

BUTTONS_FILENAME = None
EVDEV_DEVICE_NAME = None

LEDS = {}
LED_GROUPS = {}
LED_COLORS = {}
LED_DEFAULT_COLOR = ''



from mindpile.Mapping.utils import MethodCall

@MethodCall(target="Boolean_And.vix", A=bool, B=bool, Result=bool)
def booleanAnd():
    return '''
        Result = A and B
    '''

@MethodCall(target="Boolean_Or.vix", A=bool, B=bool, Result=bool)
def booleanOr():
    return '''
        Result = A or B
    '''

@MethodCall(target="Boolean_XOr.vix", A=bool, B=bool, Result=bool)
def booleanXOR():
    return '''
        Result = A != B
    '''

@MethodCall(target="Boolean_Not.vix", A=bool, Result=bool)
def booleanNot():
    return '''
        Result = not A
    '''



from abc import ABC, abstractmethod
from typing import Any

def intParser(input: str):
    return int(float(input))

def boolParser(input: str):
    return input == "True"

def get_parser(type):
    parser = None
    if type == int:
        parser = intParser
    elif type == str:
        parser = str
    elif type == bool:
        parser = boolParser
    elif type == float:
        parser = float
    elif isinstance(type, EV3Type) or issubclass(type, EV3Type):
        parser = type.parse
    return parser

def get_type(type_str: str):
    types = {
        "Int32": int,
        "String": str,
        "Boolean": bool,
        "Single": float,
        "Boolean[]": List(bool),
        "Single[]": List(float)
    }
    return types[type_str]

class Local:
    def __init__(self, initial_value) -> None:
        self.initial_value = initial_value

class EV3Type(ABC):
    @staticmethod
    @abstractmethod
    def parse(input: str) -> Any:
        pass

class List(EV3Type):
    def __init__(self, type):
        self.parser = get_parser(type)
        if self.parser is None:
            raise Exception(f"Type {type} is not a supported parameter type")
    
    def parse(self, input: str):
        parser = self.parser
        body = input[1:-1] # Remove surrounding []
        if body == '':
            items_str = []
        else:
            items_str = body.split(',')
        items = list(map(parser, items_str))
        return items

class Literal(EV3Type):
    def __init__(self, type):
        self.parser = get_parser(type)
        if self.parser is None:
            raise Exception(f"Type {type} is not a supported parameter type")
    
    def parse(self, input: str):
        return self.parser(input)

class InPort(EV3Type):
    @staticmethod
    def parse(input: str):
        # Convert "1.4" -> 4
        # TODO: EV3 allows selecting ports from other paired bricks (specified by the number)
        if len(input) > 3:
            return None
        port = input[2]
        return ord(port) - 48

class OutPort(EV3Type):
    @staticmethod
    def parse(input: str):
        # Convert "1.A" -> 1, "1.B" -> 2
        # TODO: EV3 allows selecting ports from other paired bricks (specified by the number)
        if len(input) > 3:
            return None
        port = input[2]
        return ord(port) - 64

class OutPortPair(EV3Type):
    @staticmethod
    def parse(input: str):
        # Convert "1.A+D" -> 14, "1.B+C" -> 23
        if len(input) > 5:
            return None
        portLeft = ord(input[2]) - 64
        portRight = ord(input[4]) - 64
        return portLeft * 10 + portRight



import argparse
from mindpile import from_ev3

parser = argparse.ArgumentParser(description="Transpiles Mindstorms to ev3dev2 python.")

parser.add_argument("mindstorms_path", type=str, help="The path to your mindstorms program (file ending with .ev3)")
parser.add_argument("--output_path", "-o", type=str, default="code.py", dest="output_path", help="The file we output python code to.")
parser.add_argument("--ev3sim", "-e", action="store_true", dest="ev3sim_support", help="Enable ev3sim support")

def main(passed_args=None):
    import sys
    args = parser.parse_args(passed_args or sys.argv[1:])
    python_source = from_ev3(args.mindstorms_path, ev3sim_support=args.ev3sim_support)
    # Third -> Save to the output_path.
    with open(args.output_path, "w") as f:
        f.write(python_source)
    print(f"Program written to {args.output_path}.")

if __name__ == "__main__":
    main()



# -----------------------------------------------------------------------------
# Copyright (c) 2015 Ralph Hempel <rhempel@hempeldesigngroup.com>
# Copyright (c) 2015 Anton Vanhoucke <antonvh@gmail.com>
# Copyright (c) 2015 Denis Demidov <dennis.demidov@gmail.com>
# Copyright (c) 2015 Eric Pascual <eric@pobot.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# -----------------------------------------------------------------------------

import sys
from ev3dev2.stopwatch import StopWatch
from ev3dev2 import get_current_platform, is_micropython, library_load_warning_message
from logging import getLogger

# Import the button filenames, this is platform specific
platform = get_current_platform()

if platform == 'ev3':
    from ._platform.ev3 import BUTTONS_FILENAME, EVDEV_DEVICE_NAME

elif platform == 'evb':
    from ._platform.evb import BUTTONS_FILENAME, EVDEV_DEVICE_NAME

elif platform == 'pistorms':
    from ._platform.pistorms import BUTTONS_FILENAME, EVDEV_DEVICE_NAME

elif platform == 'brickpi':
    from ._platform.brickpi import BUTTONS_FILENAME, EVDEV_DEVICE_NAME

elif platform == 'brickpi3':
    from ._platform.brickpi3 import BUTTONS_FILENAME, EVDEV_DEVICE_NAME

elif platform == 'fake':
    from ._platform.fake import BUTTONS_FILENAME, EVDEV_DEVICE_NAME

else:
    raise Exception("Unsupported platform '%s'" % platform)

if sys.version_info < (3, 4):
    raise SystemError('Must be using Python 3.4 or higher')

log = getLogger(__name__)


class MissingButton(Exception):
    pass


class ButtonCommon(object):
    def __str__(self):
        return self.__class__.__name__

    @staticmethod
    def on_change(changed_buttons):
        """
        This handler is called by ``process()`` whenever state of any button has
        changed since last ``process()`` call. ``changed_buttons`` is a list of
        tuples of changed button names and their states.
        """
        pass

    @property
    def buttons_pressed(self):
        raise NotImplementedError()

    def any(self):
        """
        Checks if any button is pressed.
        """
        return bool(self.buttons_pressed)

    def check_buttons(self, buttons=[]):
        """
        Check if currently pressed buttons exactly match the given list ``buttons``.
        """
        return set(self.buttons_pressed) == set(buttons)

    def _wait(self, wait_for_button_press, wait_for_button_release, timeout_ms):
        raise NotImplementedError()

    def wait_for_pressed(self, buttons, timeout_ms=None):
        """
        Wait for ``buttons`` to be pressed down.
        """
        return self._wait(buttons, [], timeout_ms)

    def wait_for_released(self, buttons, timeout_ms=None):
        """
        Wait for ``buttons`` to be released.
        """
        return self._wait([], buttons, timeout_ms)

    def wait_for_bump(self, buttons, timeout_ms=None):
        """
        Wait for ``buttons`` to be pressed down and then released.
        Both actions must happen within ``timeout_ms``.
        """
        stopwatch = StopWatch()
        stopwatch.start()

        if self.wait_for_pressed(buttons, timeout_ms):
            if timeout_ms is not None:
                timeout_ms -= stopwatch.value_ms
            return self.wait_for_released(buttons, timeout_ms)

        return False

    def process(self, new_state=None):
        """
        Check for currenly pressed buttons. If the ``new_state`` differs from the
        old state, call the appropriate button event handlers (on_up, on_down, etc).
        """
        if new_state is None:
            new_state = set(self.buttons_pressed)
        old_state = self._state if hasattr(self, '_state') else set()
        self._state = new_state

        state_diff = new_state.symmetric_difference(old_state)
        for button in state_diff:
            handler = getattr(self, 'on_' + button)

            if handler is not None:
                handler(button in new_state)

        if self.on_change is not None and state_diff:
            self.on_change([(button, button in new_state) for button in state_diff])


class EV3ButtonCommon(object):

    # These handlers are called by ButtonCommon.process() whenever the
    # state of 'up', 'down', etc buttons have changed since last
    # ButtonCommon.process() call
    on_up = None
    on_down = None
    on_left = None
    on_right = None
    on_enter = None
    on_backspace = None

    @property
    def up(self):
        """
        Check if ``up`` button is pressed.
        """
        return 'up' in self.buttons_pressed

    @property
    def down(self):
        """
        Check if ``down`` button is pressed.
        """
        return 'down' in self.buttons_pressed

    @property
    def left(self):
        """
        Check if ``left`` button is pressed.
        """
        return 'left' in self.buttons_pressed

    @property
    def right(self):
        """
        Check if ``right`` button is pressed.
        """
        return 'right' in self.buttons_pressed

    @property
    def enter(self):
        """
        Check if ``enter`` button is pressed.
        """
        return 'enter' in self.buttons_pressed

    @property
    def backspace(self):
        """
        Check if ``backspace`` button is pressed.
        """
        return 'backspace' in self.buttons_pressed


# micropython implementation
if is_micropython():  # noqa: C901

    try:
        # This is a linux-specific module.
        # It is required by the Button class, but failure to import it may be
        # safely ignored if one just needs to run API tests on Windows.
        import fcntl
    except ImportError:
        log.warning(library_load_warning_message("fcntl", "Button"))

    if platform not in ("ev3", "fake"):
        raise Exception("micropython button support has not been implemented for '%s'" % platform)

    def _test_bit(buf, index):
        byte = buf[int(index >> 3)]
        bit = byte & (1 << (index % 8))
        return bool(bit)

    class ButtonBase(ButtonCommon):
        pass

    class Button(ButtonCommon, EV3ButtonCommon):
        """
        EV3 Buttons
        """

        # Button key codes
        UP = 103
        DOWN = 108
        LEFT = 105
        RIGHT = 106
        ENTER = 28
        BACK = 14

        # Note, this order is intentional and comes from the EV3-G software
        _BUTTONS = (UP, DOWN, LEFT, RIGHT, ENTER, BACK)
        _BUTTON_DEV = '/dev/input/by-path/platform-gpio_keys-event'

        _BUTTON_TO_STRING = {
            UP: "up",
            DOWN: "down",
            LEFT: "left",
            RIGHT: "right",
            ENTER: "enter",
            BACK: "backspace",
        }

        # stuff from linux/input.h and linux/input-event-codes.h
        _KEY_MAX = 0x2FF
        _KEY_BUF_LEN = (_KEY_MAX + 7) // 8
        _EVIOCGKEY = 2 << (14 + 8 + 8) | _KEY_BUF_LEN << (8 + 8) | ord('E') << 8 | 0x18

        def __init__(self):
            super(Button, self).__init__()
            self._devnode = open(Button._BUTTON_DEV, 'b')
            self._fd = self._devnode.fileno()
            self._buffer = bytearray(Button._KEY_BUF_LEN)

        @property
        def buttons_pressed(self):
            """
            Returns list of pressed buttons
            """
            fcntl.ioctl(self._fd, Button._EVIOCGKEY, self._buffer, mut=True)

            pressed = []
            for b in Button._BUTTONS:
                if _test_bit(self._buffer, b):
                    pressed.append(Button._BUTTON_TO_STRING[b])
            return pressed

        def process_forever(self):
            while True:
                self.process()

        def _wait(self, wait_for_button_press, wait_for_button_release, timeout_ms):
            stopwatch = StopWatch()
            stopwatch.start()

            # wait_for_button_press/release can be a list of buttons or a string
            # with the name of a single button.  If it is a string of a single
            # button convert that to a list.
            if isinstance(wait_for_button_press, str):
                wait_for_button_press = [
                    wait_for_button_press,
                ]

            if isinstance(wait_for_button_release, str):
                wait_for_button_release = [
                    wait_for_button_release,
                ]

            while True:
                all_pressed = True
                all_released = True
                pressed = self.buttons_pressed

                for button in wait_for_button_press:
                    if button not in pressed:
                        all_pressed = False
                        break

                for button in wait_for_button_release:
                    if button in pressed:
                        all_released = False
                        break

                if all_pressed and all_released:
                    return True

                if timeout_ms is not None and stopwatch.value_ms >= timeout_ms:
                    return False


# python3 implementation
else:
    import array

    try:
        # This is a linux-specific module.
        # It is required by the Button class, but failure to import it may be
        # safely ignored if one just needs to run API tests on Windows.
        import fcntl
    except ImportError:
        log.warning(library_load_warning_message("fcntl", "Button"))

    try:
        # This is a linux-specific module.
        # It is required by the Button class, but failure to import it may be
        # safely ignored if one just needs to run API tests on Windows.
        import evdev
    except ImportError:
        log.warning(library_load_warning_message("evdev", "Button"))

    class ButtonBase(ButtonCommon):
        """
        Abstract button interface.
        """
        _state = set([])

        @property
        def evdev_device(self):
            """
            Return our corresponding evdev device object
            """
            devices = [evdev.InputDevice(fn) for fn in evdev.list_devices()]

            for device in devices:
                if device.name == self.evdev_device_name:
                    return device

            raise Exception("%s: could not find evdev device '%s'" % (self, self.evdev_device_name))

        def process_forever(self):
            for event in self.evdev_device.read_loop():
                if event.type == evdev.ecodes.EV_KEY:
                    self.process()

    class ButtonEVIO(ButtonBase):
        """
        Provides a generic button reading mechanism that works with event interface
        and may be adapted to platform specific implementations.

        This implementation depends on the availability of the EVIOCGKEY ioctl
        to be able to read the button state buffer. See Linux kernel source
        in /include/uapi/linux/input.h for details.
        """

        KEY_MAX = 0x2FF
        KEY_BUF_LEN = int((KEY_MAX + 7) / 8)
        EVIOCGKEY = (2 << (14 + 8 + 8) | KEY_BUF_LEN << (8 + 8) | ord('E') << 8 | 0x18)

        _buttons = {}

        def __init__(self):
            super(ButtonEVIO, self).__init__()
            self._file_cache = {}
            self._buffer_cache = {}

            for b in self._buttons:
                name = self._buttons[b]['name']

                if name is None:
                    raise MissingButton("Button '%s' is not available on this platform" % b)

                if name not in self._file_cache:
                    self._file_cache[name] = open(name, 'rb', 0)
                    self._buffer_cache[name] = array.array('B', [0] * self.KEY_BUF_LEN)

        def _button_file(self, name):
            return self._file_cache[name]

        def _button_buffer(self, name):
            return self._buffer_cache[name]

        @property
        def buttons_pressed(self):
            """
            Returns list of names of pressed buttons.
            """
            for b in self._buffer_cache:
                fcntl.ioctl(self._button_file(b), self.EVIOCGKEY, self._buffer_cache[b])

            pressed = []
            for k, v in self._buttons.items():
                buf = self._buffer_cache[v['name']]
                bit = v['value']

                if bool(buf[int(bit / 8)] & 1 << bit % 8):
                    pressed.append(k)

            return pressed

        def _wait(self, wait_for_button_press, wait_for_button_release, timeout_ms):
            stopwatch = StopWatch()
            stopwatch.start()

            # wait_for_button_press/release can be a list of buttons or a string
            # with the name of a single button.  If it is a string of a single
            # button convert that to a list.
            if isinstance(wait_for_button_press, str):
                wait_for_button_press = [
                    wait_for_button_press,
                ]

            if isinstance(wait_for_button_release, str):
                wait_for_button_release = [
                    wait_for_button_release,
                ]

            for event in self.evdev_device.read_loop():
                if event.type == evdev.ecodes.EV_KEY:
                    all_pressed = True
                    all_released = True
                    pressed = self.buttons_pressed

                    for button in wait_for_button_press:
                        if button not in pressed:
                            all_pressed = False
                            break

                    for button in wait_for_button_release:
                        if button in pressed:
                            all_released = False
                            break

                    if all_pressed and all_released:
                        return True

                    if timeout_ms is not None and stopwatch.value_ms >= timeout_ms:
                        return False

    class Button(ButtonEVIO, EV3ButtonCommon):
        """
        EV3 Buttons
        """

        _buttons = {
            'up': {
                'name': BUTTONS_FILENAME,
                'value': 103
            },
            'down': {
                'name': BUTTONS_FILENAME,
                'value': 108
            },
            'left': {
                'name': BUTTONS_FILENAME,
                'value': 105
            },
            'right': {
                'name': BUTTONS_FILENAME,
                'value': 106
            },
            'enter': {
                'name': BUTTONS_FILENAME,
                'value': 28
            },
            'backspace': {
                'name': BUTTONS_FILENAME,
                'value': 14
            },
        }
        evdev_device_name = EVDEV_DEVICE_NAME



import unittest
import Utility.utility  # TODO figure out imports for tests


class TestUtility(unittest.TestCase):
    def test_removeNameSpace(self):
        self.assertEqual("abc", Utility.utility.removeNameSpace("{asdfasdf}"))




# -----------------------------------------------------------------------------
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# -----------------------------------------------------------------------------

import os


class Console():
    """
    A class that represents the EV3 LCD console, which implements ANSI codes
    for cursor positioning, text color, and resetting the screen. Supports changing
    the console font using standard system fonts.
    """
    def __init__(self, font="Lat15-TerminusBold24x12"):
        """
        Construct the Console instance, optionally with a font name specified.

        Parameter:

        - ``font`` (string): Font name, as found in ``/usr/share/consolefonts/``

        """
        self._font = None
        self._columns = 0
        self._rows = 0
        self._echo = False
        self._cursor = False
        self.set_font(font, reset_console=False)  # don't reset the screen during construction
        self.cursor = False
        self.echo = False

    @property
    def columns(self):
        """
        Return (int) number of columns on the EV3 LCD console supported by the current font.
        """
        return self._columns

    @property
    def rows(self):
        """
        Return (int) number of rows on the EV3 LCD console supported by the current font.
        """
        return self._rows

    @property
    def echo(self):
        """
        Return (bool) whether the console echo mode is enabled.
        """
        return self._echo

    @echo.setter
    def echo(self, value):
        """
        Enable/disable console echo (so that EV3 button presses do not show the escape characters on
        the LCD console). Set to True to show the button codes, or False to hide them.
        """
        self._echo = value
        os.system("stty {}".format("echo" if value else "-echo"))

    @property
    def cursor(self):
        """
        Return (bool) whether the console cursor is visible.
        """
        return self._cursor

    @cursor.setter
    def cursor(self, value):
        """
        Enable/disable console cursor (to hide the cursor on the LCD).
        Set to True to show the cursor, or False to hide it.
        """
        self._cursor = value
        print("\x1b[?25{}".format('h' if value else 'l'), end='')

    def text_at(self, text, column=1, row=1, reset_console=False, inverse=False, alignment="L"):
        """
        Display ``text`` (string) at grid position (``column``, ``row``).
        Note that the grid locations are 1-based (not 0-based).

        Depending on the font, the number of columns and rows supported by the EV3 LCD console
        can vary. Large fonts support as few as 11 columns and 4 rows, while small fonts support
        44 columns and 21 rows. The default font for the Console() class results in a grid that
        is 14 columns and 5 rows.

        Using the ``inverse=True`` parameter will display the ``text`` with more emphasis and contrast,
        as the background of the text will be black, and the foreground is white. Using inverse
        can help in certain situations, such as to indicate when a color sensor senses
        black, or the gyro sensor is pointing to zero.

        Use the ``alignment`` parameter to enable the function to align the ``text`` differently to the
        column/row values passed-in. Use ``L`` for left-alignment (default), where the first character
        in the ``text`` will show at the column/row position. Use ``R`` for right-alignment, where the
        last character will show at the column/row position. Use ``C`` for center-alignment, where the
        text string will centered at the column/row position (as close as possible using integer
        division--odd-length text string will center better than even-length).

        Parameters:

        - ``text`` (string): Text to display
        - ``column`` (int): LCD column position to start the text (1 = left column);
          text will wrap when it reaches the right edge
        - ``row`` (int): LCD row position to start the text (1 = top row)
        - ``reset_console`` (bool): ``True`` to reset the EV3 LCD console before showing
          the text; default is ``False``
        - ``inverse`` (bool): ``True`` for white on black, otherwise black on white;
          default is ``False``
        - ``alignment`` (string): Align the ``text`` horizontally. Use ``L`` for left-alignment (default),
          ``R`` for right-alignment, or ``C`` for center-alignment

        """

        if reset_console:
            self.reset_console()

        if alignment == "R":
            column = column - len(text) + 1
        elif alignment == "C":
            column -= len(text) // 2

        if inverse:
            text = "\x1b[7m{}\x1b[m".format(text)

        print("\x1b[{};{}H{}".format(row, column, text), end='')

    def set_font(self, font="Lat15-TerminusBold24x12", reset_console=True):
        """
        Set the EV3 LCD console font and optionally reset the EV3 LCD console
        to clear it and turn off the cursor.

        Parameters:

        - ``font`` (string): Font name, as found in ``/usr/share/consolefonts/``
        - ``reset_console`` (bool): ``True`` to reset the EV3 LCD console
          after the font change; default is ``True``

        """
        if font is not None and font != self._font:
            self._font = font
            os.system("setfont {}".format(font))
            rows, columns = os.popen('stty size').read().strip().split(" ")
            self._rows = int(rows)
            self._columns = int(columns)

        if reset_console:
            self.reset_console()

    def clear_to_eol(self, column=None, row=None):
        """
        Clear to the end of line from the ``column`` and ``row`` position
        on the EV3 LCD console. Default to current cursor position.

        Parameters:

        - ``column`` (int): LCD column position to move to before clearing
        - ``row`` (int): LCD row position to move to before clearing

        """
        if column is not None and row is not None:
            print("\x1b[{};{}H".format(row, column), end='')
        print("\x1b[K", end='')

    def reset_console(self):
        """
        Clear the EV3 LCD console using ANSI codes, and move the cursor to 1,1
        """
        print("\x1b[2J\x1b[H", end='')



# -----------------------------------------------------------------------------
# Copyright (c) 2015 Ralph Hempel
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# -----------------------------------------------------------------------------

import sys

if sys.version_info < (3, 4):
    raise SystemError('Must be using Python 3.4 or higher')

CENTIMETER_MM = 10
DECIMETER_MM = 100
METER_MM = 1000
INCH_MM = 25.4
FOOT_MM = 304.8
YARD_MM = 914.4
STUD_MM = 8


class DistanceValue(object):
    """
    A base class for other unit types. Don't use this directly; instead, see
    :class:`DistanceMillimeters`, :class:`DistanceCentimeters`, :class:`DistanceDecimeters``, :class:`DistanceMeters`,
    :class:`DistanceInches`, :class:`DistanceFeet`, :class:`DistanceYards` and :class:`DistanceStuds`.
    """

    # This allows us to sort lists of DistanceValue objects
    def __lt__(self, other):
        return self.mm < other.mm

    def __rmul__(self, other):
        return self.__mul__(other)


class DistanceMillimeters(DistanceValue):
    """
    Distance in millimeters
    """
    def __init__(self, millimeters):
        self.millimeters = millimeters

    def __str__(self):
        return str(self.millimeters) + "mm"

    def __mul__(self, other):
        assert isinstance(other, (float, int)), "{} can only be multiplied by an int or float".format(self)
        return DistanceMillimeters(self.millimeters * other)

    @property
    def mm(self):
        return self.millimeters


class DistanceCentimeters(DistanceValue):
    """
    Distance in centimeters
    """
    def __init__(self, centimeters):
        self.centimeters = centimeters

    def __str__(self):
        return str(self.centimeters) + "cm"

    def __mul__(self, other):
        assert isinstance(other, (float, int)), "{} can only be multiplied by an int or float".format(self)
        return DistanceCentimeters(self.centimeters * other)

    @property
    def mm(self):
        return self.centimeters * CENTIMETER_MM


class DistanceDecimeters(DistanceValue):
    """
    Distance in decimeters
    """
    def __init__(self, decimeters):
        self.decimeters = decimeters

    def __str__(self):
        return str(self.decimeters) + "dm"

    def __mul__(self, other):
        assert isinstance(other, (float, int)), "{} can only be multiplied by an int or float".format(self)
        return DistanceDecimeters(self.decimeters * other)

    @property
    def mm(self):
        return self.decimeters * DECIMETER_MM


class DistanceMeters(DistanceValue):
    """
    Distance in meters
    """
    def __init__(self, meters):
        self.meters = meters

    def __str__(self):
        return str(self.meters) + "m"

    def __mul__(self, other):
        assert isinstance(other, (float, int)), "{} can only be multiplied by an int or float".format(self)
        return DistanceMeters(self.meters * other)

    @property
    def mm(self):
        return self.meters * METER_MM


class DistanceInches(DistanceValue):
    """
    Distance in inches
    """
    def __init__(self, inches):
        self.inches = inches

    def __str__(self):
        return str(self.inches) + "in"

    def __mul__(self, other):
        assert isinstance(other, (float, int)), "{} can only be multiplied by an int or float".format(self)
        return DistanceInches(self.inches * other)

    @property
    def mm(self):
        return self.inches * INCH_MM


class DistanceFeet(DistanceValue):
    """
    Distance in feet
    """
    def __init__(self, feet):
        self.feet = feet

    def __str__(self):
        return str(self.feet) + "ft"

    def __mul__(self, other):
        assert isinstance(other, (float, int)), "{} can only be multiplied by an int or float".format(self)
        return DistanceFeet(self.feet * other)

    @property
    def mm(self):
        return self.feet * FOOT_MM


class DistanceYards(DistanceValue):
    """
    Distance in yards
    """
    def __init__(self, yards):
        self.yards = yards

    def __str__(self):
        return str(self.yards) + "yd"

    def __mul__(self, other):
        assert isinstance(other, (float, int)), "{} can only be multiplied by an int or float".format(self)
        return DistanceYards(self.yards * other)

    @property
    def mm(self):
        return self.yards * YARD_MM


class DistanceStuds(DistanceValue):
    """
    Distance in studs
    """
    def __init__(self, studs):
        self.studs = studs

    def __str__(self):
        return str(self.studs) + "stud"

    def __mul__(self, other):
        assert isinstance(other, (float, int)), "{} can only be multiplied by an int or float".format(self)
        return DistanceStuds(self.studs * other)

    @property
    def mm(self):
        return self.studs * STUD_MM



import ast
from collections import OrderedDict
import copy
import functools
import textwrap
from mindpile.Utility.memo import memoise
from .types import Local, Literal, get_parser
from .boilerplate import boilerplate

methods = {}
setupCode = OrderedDict()
global_var_count = 0

def newGlobalName():
    global global_var_count
    name = f"var_{global_var_count}"
    global_var_count += 1
    return name

def newTree():
    return ast.parse("")

def substitute_tree(tree):
    class Template(ast.NodeTransformer):
        def __init__(self, substitutions={}) -> None:
            self.substitutions = substitutions

        def visit_Name(self, node: ast.Name):
            name = node.id
            if self.substitutions.get(name) is not None:
                value = self.substitutions[name]
                if isinstance(value, ast.AST):
                    return value
                else:
                    return ast.Constant(value)
            return node
        
        def run(self):
            return self.visit(copy.deepcopy(tree))
    
    def run(substitutions={}):
        template = Template(substitutions)
        return template.run()

    return run

def parse_code(code, substitutions={}):
    tree = ast.parse(textwrap.dedent(code))
    if len(substitutions) == 0:
        return tree
    sub = substitute_tree(tree)
    return sub(substitutions)
p = parse_code

def parse_parameter_types(parameters):
    types = {}
    local_variables = {}
    for name, type in parameters.items():
        if isinstance(type, Local):
            local_variables[name] = type
        else:
            parser = get_parser(type)
            if parser is None:
                raise Exception(f"Mapping parameter {name} is of an unknown type {type}")
            types[name] = type
    return types, local_variables

def create_parameter_substitutions(target, parameters, types, local_variables):
    substitutions = {}

    for name, type in types.items():
        if parameters.get(name) is None:
            raise Exception(f"Couldn't map to parameter {name} for {target}")
        value = parameters[name]
        if isinstance(value, ast.AST) and isinstance(type, Literal):
            raise Exception(f"Expected literal for parameter {name} of {target}, got variable")
        substitutions[name] = parameters[name]
    
    for name, type in local_variables.items():
        global_name = newGlobalName()
        tree = newTree()
        tree.body = [
            ast.Assign(
                targets=[ast.Name(id=global_name, ctx=ast.Store())],
                value=ast.Constant(type.initial_value)
            )
        ]
        setupCode[global_name] = tree
        substitutions[name] = ast.Name(id=global_name)
    
    return substitutions

def create_setup_code(func):
    prereqs = getattr(func, "__MINDPILE_PREREQS", None)
    if prereqs is not None:
        for p in prereqs:
            if setupCode.get(p) is None:
                tree = p()
                setupCode[p] = tree

def MethodCall(target: str, **parameters):
    def decorator(func):
        # Memoise the heavy lifting of template generation so it is only run once and only if this MethodCall is used
        @memoise
        def memo():
            types, local_variables = parse_parameter_types(parameters)

            # Get the AST template
            stringTemplate = func()
            tree = ast.parse(textwrap.dedent(stringTemplate))
            substitute = substitute_tree(tree)

            return substitute, types, local_variables
        
        @functools.wraps(func)
        def wrappedFunc(**kwargs) -> ast.AST:
            create_setup_code(func)

            substitute, types, local_variables = memo()
            substitutions = create_parameter_substitutions(target, kwargs, types, local_variables)

            for name, type in types.items():
                parser = get_parser(type)
                value = substitutions[name]
                if not isinstance(value, ast.AST):
                    substitutions[name] = ast.Constant(parser(value))

            return substitute(substitutions)
        methods[target] = wrappedFunc
        
        return func
    return decorator

def DynamicMethodCall(target: str, **parameters):
    def decorator(func):
        # Memoise the heavy lifting of template generation so it is only run once and only if this MethodCall is used
        @memoise
        def memo():
            types, local_variables = parse_parameter_types(parameters)
            return types, local_variables
    
        @functools.wraps(func)
        def wrappedFunc(**kwargs) -> ast.AST:
            create_setup_code(func)

            types, local_variables = memo()
            substitutions = create_parameter_substitutions(target, kwargs, types, local_variables)

            for name, type in types.items():
                parser = get_parser(type)
                value = substitutions[name]
                if not isinstance(value, ast.AST):
                    substitutions[name] = parser(value)

            return func(**substitutions)
        methods[target] = wrappedFunc

        return func
    return decorator

def Setup(func):
    @functools.wraps(func)
    @memoise
    def wrapped():
        stringTemplate = func()
        tree = p(stringTemplate)
        return tree
    return wrapped

def Requires(prereq):
    def decorator(func):
        prereqs = getattr(func, "__MINDPILE_PREREQS", set())
        prereqs.add(prereq)
        setattr(func, "__MINDPILE_PREREQS", prereqs)
        return func
    return decorator

def startCodeGen():
    global global_var_count
    setupCode.clear()
    global_var_count = 0

def generateSetupAST():
    tree = newTree()
    tree.body.append(p(boilerplate).body)
    for block in setupCode.values():
        tree.body += block.body
    return tree



import functools

def memoise(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        if not wrapper.hasResult:
            wrapper.result = func(*args, **kwargs)
            wrapper.hasResult = True
        return wrapper.result
    wrapper.result = None
    wrapper.hasResult = False

    return wrapper



from mindpile.Mapping.types import OutPort
from mindpile.Mapping.utils import MethodCall

@MethodCall(target="RotationValue.vix", MotorPort=OutPort, Degrees=float)
def motorDegrees():
    return '''
        Degrees = m(MotorPort).position / m(MotorPort).count_per_rot * 360
    '''

@MethodCall(target="RotationValueRotations.vix", MotorPort=OutPort, Rotations=float)
def motorRotations():
    return '''
        Degrees = m(MotorPort).position / m(MotorPort).count_per_rot
    '''

@MethodCall(target="MotorSpeedSensor.vix", MotorPort=OutPort, Speed=float)
def motorPower():
    return '''
        Speed = m(MotorPort).duty_cycle
    '''

@MethodCall(target="RotationDegreesCompare.vix", MotorPort=OutPort, Comparison=int, ThresholdDegrees=float, Result=bool, Degrees=float)
def motorDegreesCompare():
    return '''
        Degrees = m(MotorPort).position / m(MotorPort).count_per_rot * 360
        Result = compare(Comparison, Degrees, ThresholdDegrees)
    '''

@MethodCall(target="RotationRotationsCompare.vix", MotorPort=OutPort, Comparison=int, ThresholdRotations=float, Result=bool, Rotations=float)
def motorDegreesCompare():
    return '''
        Rotations = m(MotorPort).position / m(MotorPort).count_per_rot
        Result = compare(Comparison, Rotations, ThresholdRotations)
    '''

@MethodCall(target="MotorSpeedCompare.vix", MotorPort=OutPort, Comparison=int, ThresholdSpeed=float, Result=bool, Speed=float)
def motorPowerCompare():
    return '''
        Speed = m(MotorPort).duty_cycle
        Result = compare(Comparison, Speed, ThresholdSpeed)
    '''



from mindpile.Mapping.utils import MethodCall, Requires, Setup

@Setup
def ev3sim_setup():
  return '''
    import ev3sim.code_helpers
  '''

@MethodCall(target="ev3sim_wait_for_tick.vix")
@Requires(ev3sim_setup)
def wait_for_tick():
  return '''
    ev3sim.code_helpers.wait_for_tick()
  '''



# -*- coding: utf-8 -*-

# ------------------------------------------------------------------------------
# Copyright (c) 2015 Eric Pascual
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# -----------------------------------------------------------------------------
"""
An assortment of classes modeling specific features of the EV3 brick.
"""

from collections import OrderedDict

OUTPUT_A = 'ev3-ports:outA'
OUTPUT_B = 'ev3-ports:outB'
OUTPUT_C = 'ev3-ports:outC'
OUTPUT_D = 'ev3-ports:outD'

INPUT_1 = 'ev3-ports:in1'
INPUT_2 = 'ev3-ports:in2'
INPUT_3 = 'ev3-ports:in3'
INPUT_4 = 'ev3-ports:in4'

BUTTONS_FILENAME = '/dev/input/by-path/platform-gpio_keys-event'
EVDEV_DEVICE_NAME = 'EV3 Brick Buttons'

LEDS = OrderedDict()
LEDS['red_left'] = 'led0:red:brick-status'
LEDS['red_right'] = 'led1:red:brick-status'
LEDS['green_left'] = 'led0:green:brick-status'
LEDS['green_right'] = 'led1:green:brick-status'

LED_GROUPS = OrderedDict()
LED_GROUPS['LEFT'] = ('red_left', 'green_left')
LED_GROUPS['RIGHT'] = ('red_right', 'green_right')

LED_COLORS = OrderedDict()
LED_COLORS['BLACK'] = (0, 0)
LED_COLORS['RED'] = (1, 0)
LED_COLORS['GREEN'] = (0, 1)
LED_COLORS['AMBER'] = (1, 1)
LED_COLORS['ORANGE'] = (1, 0.5)
LED_COLORS['YELLOW'] = (0.1, 1)

LED_DEFAULT_COLOR = 'GREEN'



boilerplate = '''
import math
import random
import time

motors = {}
def m(port, type=None):
    if port is None:
        raise Exception("Motor block does not have a port specified")
    port = int(port)
    portName = f"out{chr(port+64)}"

    if portName not in motors:
        if type is None:
            raise Exception(f"Can't use motor {portName} without knowing its type")
        motor = type(address=portName)
        motors[portName] = motor
    return motors[portName]

sensors = {}
def s(port, type=None):
    if port is None:
        raise Exception("Sensor block does not have a port specified")
    port = int(port)
    portName = f"in{chr(port+48)}"

    if portName not in sensors:
        if type is None:
            raise Exception(f"Can't use sensor {portName} without knowing its type")
        sensor = type(address=portName)
        sensors[portName] = sensor
    return sensors[portName]

def compare(comparison, a, b):
    if comparison == 0:
        return math.isclose(a, b)
    elif comparison == 1:
        return not math.isclose(a, b)
    elif comparison == 2:
        return a > b
    elif comparison == 3:
        return a >= b
    elif comparison == 4:
        return a < b
    elif comparison == 5:
        return a <= b
    return False

def compareDirection(direction, amount, a, b):
    if direction == 0:
        return b >= a + amount
    elif direction == 1:
        return b <= a - amount
    elif direction == 2:
        return (b >= a + amount) or (b <= a - amount)
    return False
'''



from mindpile.Mapping.types import OutPort
from mindpile.Mapping.utils import MethodCall, Requires, Setup

@Setup
def largeMotorSetup():
    return '''
        from ev3dev2.motor import LargeMotor
    '''

@MethodCall(target="MotorStop.vix", MotorPort=OutPort, BrakeAtEnd=bool)
@Requires(largeMotorSetup)
def largeMotorStop():
    return '''
        m(MotorPort, LargeMotor).stop(stop_action=("hold" if BrakeAtEnd else "coast"))
    '''

@MethodCall(target="MotorUnlimited.vix", MotorPort=OutPort, Speed=float)
@Requires(largeMotorSetup)
def largeMotorUnlimited():
    return '''
        m(MotorPort, LargeMotor).on(speed=Speed)
    '''

@MethodCall(target="MotorTime.vix", MotorPort=OutPort, Speed=float, Seconds=float, BrakeAtEnd=bool)
@Requires(largeMotorSetup)
def largeMotorSeconds():
    return '''
        m(MotorPort, LargeMotor).on_for_seconds(speed=Speed, seconds=Seconds, brake=BrakeAtEnd)
    '''

@MethodCall(target="MotorDistance.vix", MotorPort=OutPort, Speed=float, Degrees=float, BrakeAtEnd=bool)
@Requires(largeMotorSetup)
def largeMotorDegrees():
    return '''
        m(MotorPort, LargeMotor).on_for_degrees(speed=Speed, degrees=Degrees, brake=BrakeAtEnd)
    '''

@MethodCall(target="MotorDistanceRotations.vix", MotorPort=OutPort, Speed=float, Rotations=float, BrakeAtEnd=bool)
@Requires(largeMotorSetup)
def largeMotorRotations():
    return '''
        m(MotorPort, LargeMotor).on_for_rotations(speed=Speed, rotations=Rotations, brake=BrakeAtEnd)
    '''



from ev3dev2 import *

platform = get_current_platform()

if platform == 'ev3':
    from ev3dev2._platform.ev3 import INPUT_1, INPUT_2, INPUT_3, INPUT_4
    from ev3dev2._platform.ev3 import OUTPUT_A, OUTPUT_B, OUTPUT_C, OUTPUT_D
    from ev3dev2._platform.ev3 import LEDS, LED_GROUPS, LED_COLORS

elif platform == 'evb':
    from ev3dev2._platform.evb import INPUT_1, INPUT_2, INPUT_3, INPUT_4
    from ev3dev2._platform.evb import OUTPUT_A, OUTPUT_B, OUTPUT_C, OUTPUT_D
    from ev3dev2._platform.evb import LEDS, LED_GROUPS, LED_COLORS

elif platform == 'pistorms':
    from ev3dev2._platform.pistorms import INPUT_1, INPUT_2, INPUT_3, INPUT_4
    from ev3dev2._platform.pistorms import OUTPUT_A, OUTPUT_B, OUTPUT_C, OUTPUT_D
    from ev3dev2._platform.pistorms import LEDS, LED_GROUPS, LED_COLORS

elif platform == 'brickpi':
    from ev3dev2._platform.brickpi import INPUT_1, INPUT_2, INPUT_3, INPUT_4
    from ev3dev2._platform.brickpi import OUTPUT_A, OUTPUT_B, OUTPUT_C, OUTPUT_D
    from ev3dev2._platform.brickpi import LEDS, LED_GROUPS, LED_COLORS

elif platform == 'brickpi3':
    from ev3dev2._platform.brickpi3 import INPUT_1, INPUT_2, INPUT_3, INPUT_4
    from ev3dev2._platform.brickpi3 import OUTPUT_A, OUTPUT_B, OUTPUT_C, OUTPUT_D
    from ev3dev2._platform.brickpi3 import LEDS, LED_GROUPS, LED_COLORS

elif platform == 'fake':
    from ev3dev2._platform.fake import INPUT_1, INPUT_2, INPUT_3, INPUT_4
    from ev3dev2._platform.fake import OUTPUT_A, OUTPUT_B, OUTPUT_C, OUTPUT_D
    from ev3dev2._platform.fake import LEDS, LED_GROUPS, LED_COLORS

else:
    raise Exception("Unsupported platform '%s'" % platform)

from ev3dev2.button import *
from ev3dev2.console import *
from ev3dev2.display import *
from ev3dev2.fonts import *
from ev3dev2.led import *
from ev3dev2.motor import *
from ev3dev2.port import *
from ev3dev2.power import *
from ev3dev2.sensor import *
from ev3dev2.sensor.lego import *
from ev3dev2.sound import *



from mindpile.Mapping.utils import MethodCall, Requires, Setup

# TODO: InvertColor, Size, and images are not supported

@Setup
def displaySetup():
    return '''
        from ev3dev2.display import Display
        import ev3dev2.fonts
        screen = Display()
        screenFont = ev3dev2.fonts.load("charB14")
    '''

@MethodCall(target="DisplayString.vix", Text=str, ClearScreen=bool, X=int, Y=int, InvertColor=bool, Size=int)
@Requires(displaySetup)
def displayStringPixel():
    return '''
        screen.text_pixels(Text, clear_screen=ClearScreen, x=X, y=Y, text_color='black', font=screenFont)
        screen.update()
    '''

@MethodCall(target="DisplayStringGrid.vix", Text=str, ClearScreen=bool, Column=int, Row=int, InvertColor=bool, Size=int)
@Requires(displaySetup)
def displayStringGrid():
    return '''
        screen.text_grid(Text, clear_screen=ClearScreen, x=Column, y=Row, text_color='black', font=screenFont)
        screen.update()
    '''

@MethodCall(target="DisplayLine.vix", ClearScreen=bool, X1=int, Y1=int, X2=int, Y2=int, InvertColor=bool)
@Requires(displaySetup)
def displayLine():
    return '''
        screen.line(clear_screen=ClearScreen, x1=X1, y1=Y1, x2=X2, y2=Y2, line_color='black', width=1)
        screen.update()
    '''

@MethodCall(target="DisplayCircle.vix", ClearScreen=bool, X=int, Y=int, Radius=int, Fill=bool, InvertColor=bool)
@Requires(displaySetup)
def displayCircle():
    return '''
        screen.circle(clear_screen=ClearScreen, x=X, y=Y, radius=Radius, fill_color=(0, 0, 0, 255 if Fill else 0), outline_color='black')
        screen.update()
    '''

@MethodCall(target="DisplayRect.vix", ClearScreen=bool, X=int, Y=int, Width=int, Height=int, Fill=bool, InvertColor=bool)
@Requires(displaySetup)
def displayRect():
    return '''
        screen.rectangle(clear_screen=ClearScreen, x1=X, y1=Y, x2=X+Width, y2=Y+Height, fill_color=(0, 0, 0, 255 if Fill else 0), outline_color='black')
        screen.update()
    '''

@MethodCall(target="DisplayPoint.vix", ClearScreen=bool, X=int, Y=int, InvertColor=bool)
@Requires(displaySetup)
def displayPoint():
    return '''
        screen.point(clear_screen=ClearScreen, x=X, y=Y, point_color='black')
        screen.update()
    '''

@Setup
def displayImageNotImplemented():
    return '''
        print("WARNING: Display Image block is not implemented")
    '''

@MethodCall(target="DisplayFile.vix", Filename=str, ClearScreen=bool, X=int, Y=int)
@Requires(displaySetup)
@Requires(displayImageNotImplemented)
def displayImage():
    return '''
    '''

@MethodCall(target="DisplayClear.vix")
@Requires(displaySetup)
def displayClear():
    return '''
        screen.clear()
    '''



import os.path
from glob import glob
from PIL import ImageFont


def available():
    """
    Returns list of available font names.
    """
    font_dir = os.path.dirname(__file__)
    names = [os.path.basename(os.path.splitext(f)[0]) for f in glob(os.path.join(font_dir, '*.pil'))]
    return sorted(names)


def load(name):
    """
    Loads the font specified by name and returns it as an instance of
    `PIL.ImageFont <http://pillow.readthedocs.io/en/latest/reference/ImageFont.html>`_
    class.
    """
    try:
        font_dir = os.path.dirname(__file__)
        pil_file = os.path.join(font_dir, '{}.pil'.format(name))
        return ImageFont.load(pil_file)
    except FileNotFoundError:
        raise Exception('Failed to load font "{}". '.format(name) +
                        'Check ev3dev.fonts.available() for the list of available fonts')



"""
An assortment of classes modeling specific features of the PiStorms.
"""
from collections import OrderedDict

OUTPUT_A = 'pistorms:BAM1'
OUTPUT_B = 'pistorms:BAM2'
OUTPUT_C = 'pistorms:BBM1'
OUTPUT_D = 'pistorms:BBM2'

INPUT_1 = 'pistorms:BAS1'
INPUT_2 = 'pistorms:BAS2'
INPUT_3 = 'pistorms:BBS1'
INPUT_4 = 'pistorms:BBS2'

BUTTONS_FILENAME = '/dev/input/by-path/platform-3f804000.i2c-event'
EVDEV_DEVICE_NAME = 'PiStorms'

LEDS = OrderedDict()
LEDS['red_left'] = 'pistorms:BB:red:brick-status'
LEDS['red_right'] = 'pistorms:BA:red:brick-status'
LEDS['green_left'] = 'pistorms:BB:green:brick-status'
LEDS['green_right'] = 'pistorms:BA:green:brick-status'
LEDS['blue_left'] = 'pistorms:BB:blue:brick-status'
LEDS['blue_right'] = 'pistorms:BA:blue:brick-status'

LED_GROUPS = OrderedDict()
LED_GROUPS['LEFT'] = ('red_left', 'green_left', 'blue_left')
LED_GROUPS['RIGHT'] = ('red_right', 'green_right', 'blue_right')

LED_COLORS = OrderedDict()
LED_COLORS['BLACK'] = (0, 0, 0)
LED_COLORS['RED'] = (1, 0, 0)
LED_COLORS['GREEN'] = (0, 1, 0)
LED_COLORS['BLUE'] = (0, 0, 1)
LED_COLORS['YELLOW'] = (1, 1, 0)
LED_COLORS['CYAN'] = (0, 1, 1)
LED_COLORS['MAGENTA'] = (1, 0, 1)

LED_DEFAULT_COLOR = 'GREEN'



from mindpile.Mapping.types import OutPortPair
from mindpile.Mapping.utils import MethodCall, Requires, Setup

@Setup
def steeringSetup():
    return '''
        from ev3dev2.motor import MoveSteering

        steeringControllers = {}
        def steering(port):
            if port is None:
                raise Exception("Steering block does not have a port specified")
            port = int(port)

            controller = steeringControllers.get(port)
            if controller is None:
                portLeft = f"out{chr((port // 10) + 64)}"
                portRight = f"out{chr((port % 10) + 64)}"
                controller = MoveSteering(portLeft, portRight)
                steeringControllers[port] = controller
            return controller
    '''

@MethodCall(target="MoveStop.vix", Ports=OutPortPair, BrakeAtEnd=bool)
@Requires(steeringSetup)
def steeringStop():
    return '''
        steering(Ports).off(brake=BrakeAtEnd)
    '''

@MethodCall(target="MoveUnlimited.vix", Ports=OutPortPair, Steering=float, Speed=float)
@Requires(steeringSetup)
def steeringUnlimited():
    return '''
        steering(Ports).on(steering=Steering, speed=Speed)
    '''

@MethodCall(target="MoveTime.vix", Ports=OutPortPair, Steering=float, Speed=float, Seconds=float, BrakeAtEnd=bool)
@Requires(steeringSetup)
def steeringSeconds():
    return '''
        steering(Ports).on_for_seconds(steering=Steering, speed=Speed, seconds=Seconds, brake=BrakeAtEnd)
    '''

@MethodCall(target="MoveDistance.vix", Ports=OutPortPair, Steering=float, Speed=float, Degrees=float, BrakeAtEnd=bool)
@Requires(steeringSetup)
def steeringDegrees():
    return '''
        steering(Ports).on_for_degrees(steering=Steering, speed=Speed, degrees=Degrees, brake=BrakeAtEnd)
    '''

@MethodCall(target="MoveDistanceRotations.vix", Ports=OutPortPair, Steering=float, Speed=float, Rotations=float, BrakeAtEnd=bool)
@Requires(steeringSetup)
def steeringRotations():
    return '''
        steering(Ports).on_for_rotations(steering=Steering, speed=Speed, rotations=Rotations, brake=BrakeAtEnd)
    '''



# -----------------------------------------------------------------------------
# Copyright (c) 2015 Ralph Hempel <rhempel@hempeldesigngroup.com>
# Copyright (c) 2015 Anton Vanhoucke <antonvh@gmail.com>
# Copyright (c) 2015 Denis Demidov <dennis.demidov@gmail.com>
# Copyright (c) 2015 Eric Pascual <eric@pobot.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# -----------------------------------------------------------------------------

import sys
import os
import stat
import time
import _thread
from collections import OrderedDict
from ev3dev2 import get_current_platform, Device
from ev3dev2.stopwatch import StopWatch
from time import sleep

if sys.version_info < (3, 4):
    raise SystemError('Must be using Python 3.4 or higher')

# Import the LED settings, this is platform specific
platform = get_current_platform()

if platform == 'ev3':
    from ev3dev2._platform.ev3 import LEDS, LED_GROUPS, LED_COLORS, LED_DEFAULT_COLOR

elif platform == 'evb':
    from ev3dev2._platform.evb import LEDS, LED_GROUPS, LED_COLORS, LED_DEFAULT_COLOR

elif platform == 'pistorms':
    from ev3dev2._platform.pistorms import LEDS, LED_GROUPS, LED_COLORS, LED_DEFAULT_COLOR

elif platform == 'brickpi':
    from ev3dev2._platform.brickpi import LEDS, LED_GROUPS, LED_COLORS, LED_DEFAULT_COLOR

elif platform == 'brickpi3':
    from ev3dev2._platform.brickpi3 import LEDS, LED_GROUPS, LED_COLORS, LED_DEFAULT_COLOR

elif platform == 'fake':
    from ev3dev2._platform.fake import LEDS, LED_GROUPS, LED_COLORS, LED_DEFAULT_COLOR

else:
    raise Exception("Unsupported platform '%s'" % platform)


class Led(Device):
    """
    Any device controlled by the generic LED driver.
    See https://www.kernel.org/doc/Documentation/leds/leds-class.txt
    for more details.
    """

    SYSTEM_CLASS_NAME = 'leds'
    SYSTEM_DEVICE_NAME_CONVENTION = '*'
    __slots__ = [
        '_max_brightness',
        '_brightness',
        '_triggers',
        '_trigger',
        '_delay_on',
        '_delay_off',
        'desc',
    ]

    def __init__(self, name_pattern=SYSTEM_DEVICE_NAME_CONVENTION, name_exact=False, desc=None, **kwargs):
        self.desc = desc
        super(Led, self).__init__(self.SYSTEM_CLASS_NAME, name_pattern, name_exact, **kwargs)
        self._max_brightness = None
        self._brightness = None
        self._triggers = None
        self._trigger = None
        self._delay_on = None
        self._delay_off = None

    def __str__(self):
        if self.desc:
            return self.desc
        else:
            return Device.__str__(self)

    @property
    def max_brightness(self):
        """
        Returns the maximum allowable brightness value.
        """
        self._max_brightness, value = self.get_cached_attr_int(self._max_brightness, 'max_brightness')
        return value

    @property
    def brightness(self):
        """
        Sets the brightness level. Possible values are from 0 to ``max_brightness``.
        """
        self._brightness, value = self.get_attr_int(self._brightness, 'brightness')
        return value

    @brightness.setter
    def brightness(self, value):
        self._brightness = self.set_attr_int(self._brightness, 'brightness', value)

    @property
    def triggers(self):
        """
        Returns a list of available triggers.
        """
        self._triggers, value = self.get_attr_set(self._triggers, 'trigger')
        return value

    @property
    def trigger(self):
        """
        Sets the LED trigger. A trigger is a kernel based source of LED events.
        Triggers can either be simple or complex. A simple trigger isn't
        configurable and is designed to slot into existing subsystems with
        minimal additional code. Examples are the ``ide-disk`` and ``nand-disk``
        triggers.

        Complex triggers whilst available to all LEDs have LED specific
        parameters and work on a per LED basis. The ``timer`` trigger is an example.
        The ``timer`` trigger will periodically change the LED brightness between
        0 and the current brightness setting. The ``on`` and ``off`` time can
        be specified via ``delay_{on,off}`` attributes in milliseconds.
        You can change the brightness value of a LED independently of the timer
        trigger. However, if you set the brightness value to 0 it will
        also disable the ``timer`` trigger.
        """
        self._trigger, value = self.get_attr_from_set(self._trigger, 'trigger')
        return value

    @trigger.setter
    def trigger(self, value):
        self._trigger = self.set_attr_string(self._trigger, 'trigger', value)

        # Workaround for ev3dev/ev3dev#225.
        # When trigger is set to 'timer', we need to wait for 'delay_on' and
        # 'delay_off' attributes to appear with correct permissions.
        if value == 'timer':
            for attr in ('delay_on', 'delay_off'):
                path = self._path + '/' + attr

                # Make sure the file has been created:
                for _ in range(5):
                    if os.path.exists(path):
                        break
                    time.sleep(0.2)
                else:
                    raise Exception('"{}" attribute has not been created'.format(attr))

                # Make sure the file has correct permissions:
                for _ in range(5):
                    mode = stat.S_IMODE(os.stat(path)[stat.ST_MODE])
                    if mode & stat.S_IRGRP and mode & stat.S_IWGRP:
                        break
                    time.sleep(0.2)
                else:
                    raise Exception('"{}" attribute has wrong permissions'.format(attr))

    @property
    def delay_on(self):
        """
        The ``timer`` trigger will periodically change the LED brightness between
        0 and the current brightness setting. The ``on`` time can
        be specified via ``delay_on`` attribute in milliseconds.
        """

        # Workaround for ev3dev/ev3dev#225.
        # 'delay_on' and 'delay_off' attributes are created when trigger is set
        # to 'timer', and destroyed when it is set to anything else.
        # This means the file cache may become outdated, and we may have to
        # reopen the file.
        for retry in (True, False):
            try:
                self._delay_on, value = self.get_attr_int(self._delay_on, 'delay_on')
                return value
            except OSError:
                if retry:
                    self._delay_on = None
                else:
                    raise

    @delay_on.setter
    def delay_on(self, value):
        # Workaround for ev3dev/ev3dev#225.
        # 'delay_on' and 'delay_off' attributes are created when trigger is set
        # to 'timer', and destroyed when it is set to anything else.
        # This means the file cache may become outdated, and we may have to
        # reopen the file.
        for retry in (True, False):
            try:
                self._delay_on = self.set_attr_int(self._delay_on, 'delay_on', value)
                return
            except OSError:
                if retry:
                    self._delay_on = None
                else:
                    raise

    @property
    def delay_off(self):
        """
        The ``timer`` trigger will periodically change the LED brightness between
        0 and the current brightness setting. The ``off`` time can
        be specified via ``delay_off`` attribute in milliseconds.
        """

        # Workaround for ev3dev/ev3dev#225.
        # 'delay_on' and 'delay_off' attributes are created when trigger is set
        # to 'timer', and destroyed when it is set to anything else.
        # This means the file cache may become outdated, and we may have to
        # reopen the file.
        for retry in (True, False):
            try:
                self._delay_off, value = self.get_attr_int(self._delay_off, 'delay_off')
                return value
            except OSError:
                if retry:
                    self._delay_off = None
                else:
                    raise

    @delay_off.setter
    def delay_off(self, value):
        """
        Workaround for ev3dev/ev3dev#225.
        ``delay_on`` and ``delay_off`` attributes are created when trigger is set
        to ``timer``, and destroyed when it is set to anything else.
        This means the file cache may become outdated, and we may have to
        reopen the file.
        """
        for retry in (True, False):
            try:
                self._delay_off = self.set_attr_int(self._delay_off, 'delay_off', value)
                return
            except OSError:
                if retry:
                    self._delay_off = None
                else:
                    raise

    @property
    def brightness_pct(self):
        """
        Returns LED brightness as a fraction of max_brightness
        """
        return float(self.brightness) / self.max_brightness

    @brightness_pct.setter
    def brightness_pct(self, value):
        self.brightness = value * self.max_brightness


class Leds(object):
    def __init__(self):
        self.leds = OrderedDict()
        self.led_groups = OrderedDict()
        self.led_colors = LED_COLORS
        self.animate_thread_id = None
        self.animate_thread_stop = False

        for (key, value) in LEDS.items():
            self.leds[key] = Led(name_pattern=value, desc=key)

        for (key, value) in LED_GROUPS.items():
            self.led_groups[key] = []

            for led_name in value:
                self.led_groups[key].append(self.leds[led_name])

    def __str__(self):
        return self.__class__.__name__

    def set_color(self, group, color, pct=1):
        """
        Sets brightness of LEDs in the given group to the values specified in
        color tuple. When percentage is specified, brightness of each LED is
        reduced proportionally.

        Example::

            my_leds = Leds()
            my_leds.set_color('LEFT', 'AMBER')

        With a custom color::

            my_leds = Leds()
            my_leds.set_color('LEFT', (0.5, 0.3))
        """
        # If this is a platform without LEDs there is nothing to do
        if not self.leds:
            return

        color_tuple = color
        if isinstance(color, str):
            assert color in self.led_colors, \
                "%s is an invalid LED color, valid choices are %s" % \
                (color, ', '.join(self.led_colors.keys()))
            color_tuple = self.led_colors[color]

        assert group in self.led_groups, \
            "%s is an invalid LED group, valid choices are %s" % \
            (group, ', '.join(self.led_groups.keys()))

        for led, value in zip(self.led_groups[group], color_tuple):
            led.brightness_pct = value * pct

    def set(self, group, **kwargs):
        """
        Set attributes for each LED in group.

        Example::

            my_leds = Leds()
            my_leds.set_color('LEFT', brightness_pct=0.5, trigger='timer')
        """

        # If this is a platform without LEDs there is nothing to do
        if not self.leds:
            return

        assert group in self.led_groups, \
            "%s is an invalid LED group, valid choices are %s" % \
            (group, ', '.join(self.led_groups.keys()))

        for led in self.led_groups[group]:
            for k in kwargs:
                setattr(led, k, kwargs[k])

    def all_off(self):
        """
        Turn all LEDs off
        """

        # If this is a platform without LEDs there is nothing to do
        if not self.leds:
            return

        self.animate_stop()

        for led in self.leds.values():
            led.brightness = 0

    def reset(self):
        """
        Put all LEDs back to their default color
        """

        if not self.leds:
            return

        self.animate_stop()

        for group in self.led_groups:
            self.set_color(group, LED_DEFAULT_COLOR)

    def animate_stop(self):
        """
        Signal the current animation thread to exit and wait for it to exit
        """

        if self.animate_thread_id:
            self.animate_thread_stop = True

            while self.animate_thread_id:
                pass

    def animate_police_lights(self,
                              color1,
                              color2,
                              group1='LEFT',
                              group2='RIGHT',
                              sleeptime=0.5,
                              duration=5,
                              block=True):
        """
        Cycle the ``group1`` and ``group2`` LEDs between ``color1`` and ``color2``
        to give the effect of police lights.  Alternate the ``group1`` and ``group2``
        LEDs every ``sleeptime`` seconds.

        Animate for ``duration`` seconds.  If ``duration`` is None animate for forever.

        Example:

        .. code-block:: python

            from ev3dev2.led import Leds
            leds = Leds()
            leds.animate_police_lights('RED', 'GREEN', sleeptime=0.75, duration=10)
        """
        def _animate_police_lights():
            self.all_off()
            even = True
            duration_ms = duration * 1000 if duration is not None else None
            stopwatch = StopWatch()
            stopwatch.start()

            while True:
                if even:
                    self.set_color(group1, color1)
                    self.set_color(group2, color2)
                else:
                    self.set_color(group1, color2)
                    self.set_color(group2, color1)

                if self.animate_thread_stop or stopwatch.is_elapsed_ms(duration_ms):
                    break

                even = not even
                sleep(sleeptime)

            self.animate_thread_stop = False
            self.animate_thread_id = None

        self.animate_stop()

        if block:
            _animate_police_lights()
        else:
            self.animate_thread_id = _thread.start_new_thread(_animate_police_lights, ())

    def animate_flash(self, color, groups=('LEFT', 'RIGHT'), sleeptime=0.5, duration=5, block=True):
        """
        Turn all LEDs in ``groups`` off/on to ``color`` every ``sleeptime`` seconds

        Animate for ``duration`` seconds.  If ``duration`` is None animate for forever.

        Example:

        .. code-block:: python

            from ev3dev2.led import Leds
            leds = Leds()
            leds.animate_flash('AMBER', sleeptime=0.75, duration=10)
        """
        def _animate_flash():
            even = True
            duration_ms = duration * 1000 if duration is not None else None
            stopwatch = StopWatch()
            stopwatch.start()

            while True:
                if even:
                    for group in groups:
                        self.set_color(group, color)
                else:
                    self.all_off()

                if self.animate_thread_stop or stopwatch.is_elapsed_ms(duration_ms):
                    break

                even = not even
                sleep(sleeptime)

            self.animate_thread_stop = False
            self.animate_thread_id = None

        self.animate_stop()

        if block:
            _animate_flash()
        else:
            self.animate_thread_id = _thread.start_new_thread(_animate_flash, ())

    def animate_cycle(self, colors, groups=('LEFT', 'RIGHT'), sleeptime=0.5, duration=5, block=True):
        """
        Cycle ``groups`` LEDs through ``colors``. Do this in a loop where
        we display each color for ``sleeptime`` seconds.

        Animate for ``duration`` seconds.  If ``duration`` is None animate for forever.

        Example:

        .. code-block:: python

            from ev3dev2.led import Leds
            leds = Leds()
            leds.animate_cyle(('RED', 'GREEN', 'AMBER'))
        """
        def _animate_cycle():
            index = 0
            max_index = len(colors)
            duration_ms = duration * 1000 if duration is not None else None
            stopwatch = StopWatch()
            stopwatch.start()

            while True:
                for group in groups:
                    self.set_color(group, colors[index])

                index += 1

                if index == max_index:
                    index = 0

                if self.animate_thread_stop or stopwatch.is_elapsed_ms(duration_ms):
                    break

                sleep(sleeptime)

            self.animate_thread_stop = False
            self.animate_thread_id = None

        self.animate_stop()

        if block:
            _animate_cycle()
        else:
            self.animate_thread_id = _thread.start_new_thread(_animate_cycle, ())

    def animate_rainbow(self, group1='LEFT', group2='RIGHT', increment_by=0.1, sleeptime=0.1, duration=5, block=True):
        """
        Gradually fade from one color to the next

        Animate for ``duration`` seconds.  If ``duration`` is None animate for forever.

        Example:

        .. code-block:: python

            from ev3dev2.led import Leds
            leds = Leds()
            leds.animate_rainbow()
        """
        def _animate_rainbow():
            # state 0: (LEFT,RIGHT) from (0,0) to (1,0)...RED
            # state 1: (LEFT,RIGHT) from (1,0) to (1,1)...AMBER
            # state 2: (LEFT,RIGHT) from (1,1) to (0,1)...GREEN
            # state 3: (LEFT,RIGHT) from (0,1) to (0,0)...OFF
            state = 0
            left_value = 0
            right_value = 0
            MIN_VALUE = 0
            MAX_VALUE = 1
            self.all_off()
            duration_ms = duration * 1000 if duration is not None else None
            stopwatch = StopWatch()
            stopwatch.start()

            while True:

                if state == 0:
                    left_value += increment_by
                elif state == 1:
                    right_value += increment_by
                elif state == 2:
                    left_value -= increment_by
                elif state == 3:
                    right_value -= increment_by
                else:
                    raise Exception("Invalid state {}".format(state))

                # Keep left_value and right_value within the MIN/MAX values
                left_value = min(left_value, MAX_VALUE)
                right_value = min(right_value, MAX_VALUE)
                left_value = max(left_value, MIN_VALUE)
                right_value = max(right_value, MIN_VALUE)

                self.set_color(group1, (left_value, right_value))
                self.set_color(group2, (left_value, right_value))

                if state == 0 and left_value == MAX_VALUE:
                    state = 1
                elif state == 1 and right_value == MAX_VALUE:
                    state = 2
                elif state == 2 and left_value == MIN_VALUE:
                    state = 3
                elif state == 3 and right_value == MIN_VALUE:
                    state = 0

                if self.animate_thread_stop or stopwatch.is_elapsed_ms(duration_ms):
                    break

                sleep(sleeptime)

            self.animate_thread_stop = False
            self.animate_thread_id = None

        self.animate_stop()

        if block:
            _animate_rainbow()
        else:
            self.animate_thread_id = _thread.start_new_thread(_animate_rainbow, ())



# -----------------------------------------------------------------------------
# Copyright (c) 2015 Ralph Hempel <rhempel@hempeldesigngroup.com>
# Copyright (c) 2015 Anton Vanhoucke <antonvh@gmail.com>
# Copyright (c) 2015 Denis Demidov <dennis.demidov@gmail.com>
# Copyright (c) 2015 Eric Pascual <eric@pobot.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# -----------------------------------------------------------------------------

import sys
import logging
import time
from ev3dev2.button import ButtonBase
from ev3dev2.sensor import Sensor

if sys.version_info < (3, 4):
    raise SystemError('Must be using Python 3.4 or higher')

log = logging.getLogger(__name__)


class TouchSensor(Sensor):
    """
    Touch Sensor
    """

    SYSTEM_CLASS_NAME = Sensor.SYSTEM_CLASS_NAME
    SYSTEM_DEVICE_NAME_CONVENTION = Sensor.SYSTEM_DEVICE_NAME_CONVENTION

    #: Button state
    MODE_TOUCH = 'TOUCH'
    MODES = (MODE_TOUCH, )

    def __init__(self, address=None, name_pattern=SYSTEM_DEVICE_NAME_CONVENTION, name_exact=False, **kwargs):
        super(TouchSensor, self).__init__(address,
                                          name_pattern,
                                          name_exact,
                                          driver_name=['lego-ev3-touch', 'lego-nxt-touch'],
                                          **kwargs)

    @property
    def is_pressed(self):
        """
        A boolean indicating whether the current touch sensor is being
        pressed.
        """
        self._ensure_mode(self.MODE_TOUCH)
        return self.value(0)

    @property
    def is_released(self):
        return not self.is_pressed

    def _wait(self, wait_for_press, timeout_ms, sleep_ms):
        tic = time.time()

        if sleep_ms:
            sleep_ms = float(sleep_ms / 1000)

        # The kernel does not supoort POLLPRI or POLLIN for sensors so we have
        # to drop into a loop and check often
        while True:

            if self.is_pressed == wait_for_press:
                return True

            if timeout_ms is not None and time.time() >= tic + timeout_ms / 1000:
                return False

            if sleep_ms:
                time.sleep(sleep_ms)

    def wait_for_pressed(self, timeout_ms=None, sleep_ms=10):
        """
        Wait for the touch sensor to be pressed down.
        """
        return self._wait(True, timeout_ms, sleep_ms)

    def wait_for_released(self, timeout_ms=None, sleep_ms=10):
        """
        Wait for the touch sensor to be released.
        """
        return self._wait(False, timeout_ms, sleep_ms)

    def wait_for_bump(self, timeout_ms=None, sleep_ms=10):
        """
        Wait for the touch sensor to be pressed down and then released.
        Both actions must happen within timeout_ms.
        """
        start_time = time.time()

        if self.wait_for_pressed(timeout_ms, sleep_ms):
            if timeout_ms is not None:
                timeout_ms -= int((time.time() - start_time) * 1000)
            return self.wait_for_released(timeout_ms, sleep_ms)

        return False


class ColorSensor(Sensor):
    """
    LEGO EV3 color sensor.
    """

    __slots__ = ['red_max', 'green_max', 'blue_max']

    SYSTEM_CLASS_NAME = Sensor.SYSTEM_CLASS_NAME
    SYSTEM_DEVICE_NAME_CONVENTION = Sensor.SYSTEM_DEVICE_NAME_CONVENTION

    #: Reflected light. Red LED on.
    MODE_COL_REFLECT = 'COL-REFLECT'

    #: Ambient light. Blue LEDs on.
    MODE_COL_AMBIENT = 'COL-AMBIENT'

    #: Color. All LEDs rapidly cycling, appears white.
    MODE_COL_COLOR = 'COL-COLOR'

    #: Raw reflected. Red LED on
    MODE_REF_RAW = 'REF-RAW'

    #: Raw Color Components. All LEDs rapidly cycling, appears white.
    MODE_RGB_RAW = 'RGB-RAW'

    #: No color.
    COLOR_NOCOLOR = 0

    #: Black color.
    COLOR_BLACK = 1

    #: Blue color.
    COLOR_BLUE = 2

    #: Green color.
    COLOR_GREEN = 3

    #: Yellow color.
    COLOR_YELLOW = 4

    #: Red color.
    COLOR_RED = 5

    #: White color.
    COLOR_WHITE = 6

    #: Brown color.
    COLOR_BROWN = 7

    MODES = (MODE_COL_REFLECT, MODE_COL_AMBIENT, MODE_COL_COLOR, MODE_REF_RAW, MODE_RGB_RAW)

    COLORS = (
        'NoColor',
        'Black',
        'Blue',
        'Green',
        'Yellow',
        'Red',
        'White',
        'Brown',
    )

    def __init__(self, address=None, name_pattern=SYSTEM_DEVICE_NAME_CONVENTION, name_exact=False, **kwargs):
        super(ColorSensor, self).__init__(address, name_pattern, name_exact, driver_name='lego-ev3-color', **kwargs)

        # See calibrate_white() for more details
        self.red_max = 300
        self.green_max = 300
        self.blue_max = 300

    @property
    def reflected_light_intensity(self):
        """
        Reflected light intensity as a percentage (0 to 100). Light on sensor is red.
        """
        self._ensure_mode(self.MODE_COL_REFLECT)
        return self.value(0)

    @property
    def ambient_light_intensity(self):
        """
        Ambient light intensity, as a percentage (0 to 100). Light on sensor is dimly lit blue.
        """
        self._ensure_mode(self.MODE_COL_AMBIENT)
        return self.value(0)

    @property
    def color(self):
        """
        Color detected by the sensor, categorized by overall value.
          - 0: No color
          - 1: Black
          - 2: Blue
          - 3: Green
          - 4: Yellow
          - 5: Red
          - 6: White
          - 7: Brown
        """
        self._ensure_mode(self.MODE_COL_COLOR)
        return self.value(0)

    @property
    def color_name(self):
        """
        Returns NoColor, Black, Blue, etc
        """
        return self.COLORS[self.color]

    @property
    def raw(self):
        """
        Red, green, and blue components of the detected color, as a tuple.

        Officially in the range 0-1020 but the values returned will never be
        that high. We do not yet know why the values returned are low, but
        pointing the color sensor at a well lit sheet of white paper will return
        values in the 250-400 range.

        If this is an issue, check out the rgb() and calibrate_white() methods.
        """
        self._ensure_mode(self.MODE_RGB_RAW)
        return self.value(0), self.value(1), self.value(2)

    def calibrate_white(self):
        """
        The RGB raw values are on a scale of 0-1020 but you never see a value
        anywhere close to 1020.  This function is designed to be called when
        the sensor is placed over a white object in order to figure out what
        are the maximum RGB values the robot can expect to see.  We will use
        these maximum values to scale future raw values to a 0-255 range in
        rgb().

        If you never call this function red_max, green_max, and blue_max will
        use a default value of 300.  This default was selected by measuring
        the RGB values of a white sheet of paper in a well lit room.

        Note that there are several variables that influence the maximum RGB
        values detected by the color sensor
        - the distance of the color sensor to the white object
        - the amount of light in the room
        - shadows that the robot casts on the sensor
        """
        (self.red_max, self.green_max, self.blue_max) = self.raw

    @property
    def rgb(self):
        """
        Same as raw() but RGB values are scaled to 0-255
        """
        (red, green, blue) = self.raw

        return (min(int((red * 255) / self.red_max), 255), min(int((green * 255) / self.green_max),
                                                               255), min(int((blue * 255) / self.blue_max), 255))

    @property
    def lab(self):
        """
        Return colors in Lab color space
        """
        RGB = [0, 0, 0]
        XYZ = [0, 0, 0]

        for (num, value) in enumerate(self.rgb):
            if value > 0.04045:
                value = pow(((value + 0.055) / 1.055), 2.4)
            else:
                value = value / 12.92

            RGB[num] = value * 100.0

        # http://www.brucelindbloom.com/index.html?Eqn_RGB_XYZ_Matrix.html
        # sRGB
        # 0.4124564  0.3575761  0.1804375
        # 0.2126729  0.7151522  0.0721750
        # 0.0193339  0.1191920  0.9503041
        X = (RGB[0] * 0.4124564) + (RGB[1] * 0.3575761) + (RGB[2] * 0.1804375)
        Y = (RGB[0] * 0.2126729) + (RGB[1] * 0.7151522) + (RGB[2] * 0.0721750)
        Z = (RGB[0] * 0.0193339) + (RGB[1] * 0.1191920) + (RGB[2] * 0.9503041)

        XYZ[0] = X / 95.047  # ref_X =  95.047
        XYZ[1] = Y / 100.0  # ref_Y = 100.000
        XYZ[2] = Z / 108.883  # ref_Z = 108.883

        for (num, value) in enumerate(XYZ):
            if value > 0.008856:
                value = pow(value, (1.0 / 3.0))
            else:
                value = (7.787 * value) + (16 / 116.0)

            XYZ[num] = value

        L = (116.0 * XYZ[1]) - 16
        a = 500.0 * (XYZ[0] - XYZ[1])
        b = 200.0 * (XYZ[1] - XYZ[2])

        L = round(L, 4)
        a = round(a, 4)
        b = round(b, 4)

        return (L, a, b)

    @property
    def hsv(self):
        """
        HSV: Hue, Saturation, Value
        H: position in the spectrum
        S: color saturation ("purity")
        V: color brightness
        """
        (r, g, b) = self.rgb
        maxc = max(r, g, b)
        minc = min(r, g, b)
        v = maxc

        if minc == maxc:
            return 0.0, 0.0, v

        s = (maxc - minc) / maxc
        rc = (maxc - r) / (maxc - minc)
        gc = (maxc - g) / (maxc - minc)
        bc = (maxc - b) / (maxc - minc)

        if r == maxc:
            h = bc - gc
        elif g == maxc:
            h = 2.0 + rc - bc
        else:
            h = 4.0 + gc - rc

        h = (h / 6.0) % 1.0

        return (h, s, v)

    @property
    def hls(self):
        """
        HLS: Hue, Luminance, Saturation
        H: position in the spectrum
        L: color lightness
        S: color saturation
        """
        (red, green, blue) = self.rgb
        maxc = max(red, green, blue)
        minc = min(red, green, blue)
        luminance = (minc + maxc) / 2.0

        if minc == maxc:
            return 0.0, luminance, 0.0

        if luminance <= 0.5:
            saturation = (maxc - minc) / (maxc + minc)
        else:
            if 2.0 - maxc - minc == 0:
                saturation = 0
            else:
                saturation = (maxc - minc) / (2.0 - maxc - minc)

        rc = (maxc - red) / (maxc - minc)
        gc = (maxc - green) / (maxc - minc)
        bc = (maxc - blue) / (maxc - minc)

        if red == maxc:
            hue = bc - gc
        elif green == maxc:
            hue = 2.0 + rc - bc
        else:
            hue = 4.0 + gc - rc

        hue = (hue / 6.0) % 1.0

        return (hue, luminance, saturation)

    @property
    def red(self):
        """
        Red component of the detected color, in the range 0-1020.
        """
        self._ensure_mode(self.MODE_RGB_RAW)
        return self.value(0)

    @property
    def green(self):
        """
        Green component of the detected color, in the range 0-1020.
        """
        self._ensure_mode(self.MODE_RGB_RAW)
        return self.value(1)

    @property
    def blue(self):
        """
        Blue component of the detected color, in the range 0-1020.
        """
        self._ensure_mode(self.MODE_RGB_RAW)
        return self.value(2)


class UltrasonicSensor(Sensor):
    """
    LEGO EV3 ultrasonic sensor.
    """

    SYSTEM_CLASS_NAME = Sensor.SYSTEM_CLASS_NAME
    SYSTEM_DEVICE_NAME_CONVENTION = Sensor.SYSTEM_DEVICE_NAME_CONVENTION

    #: Continuous measurement in centimeters.
    MODE_US_DIST_CM = 'US-DIST-CM'

    #: Continuous measurement in inches.
    MODE_US_DIST_IN = 'US-DIST-IN'

    #: Listen.
    MODE_US_LISTEN = 'US-LISTEN'

    #: Single measurement in centimeters.
    MODE_US_SI_CM = 'US-SI-CM'

    #: Single measurement in inches.
    MODE_US_SI_IN = 'US-SI-IN'

    MODES = (
        MODE_US_DIST_CM,
        MODE_US_DIST_IN,
        MODE_US_LISTEN,
        MODE_US_SI_CM,
        MODE_US_SI_IN,
    )

    def __init__(self, address=None, name_pattern=SYSTEM_DEVICE_NAME_CONVENTION, name_exact=False, **kwargs):
        super(UltrasonicSensor, self).__init__(address,
                                               name_pattern,
                                               name_exact,
                                               driver_name=['lego-ev3-us', 'lego-nxt-us'],
                                               **kwargs)

    @property
    def distance_centimeters_continuous(self):
        """
        Measurement of the distance detected by the sensor,
        in centimeters.

        The sensor will continue to take measurements so
        they are available for future reads.

        Prefer using the equivalent :meth:`UltrasonicSensor.distance_centimeters` property.
        """
        self._ensure_mode(self.MODE_US_DIST_CM)
        return self.value(0) * self._scale('US_DIST_CM')

    @property
    def distance_centimeters_ping(self):
        """
        Measurement of the distance detected by the sensor,
        in centimeters.

        The sensor will take a single measurement then stop
        broadcasting.

        If you use this property too frequently (e.g. every
        100msec), the sensor will sometimes lock up and writing
        to the mode attribute will return an error. A delay of
        250msec between each usage seems sufficient to keep the
        sensor from locking up.
        """
        # This mode is special; setting the mode causes the sensor to send out
        # a "ping", but the mode isn't actually changed.
        self.mode = self.MODE_US_SI_CM
        return self.value(0) * self._scale('US_DIST_CM')

    @property
    def distance_centimeters(self):
        """
        Measurement of the distance detected by the sensor,
        in centimeters.

        Equivalent to :meth:`UltrasonicSensor.distance_centimeters_continuous`.
        """
        return self.distance_centimeters_continuous

    @property
    def distance_inches_continuous(self):
        """
        Measurement of the distance detected by the sensor,
        in inches.

        The sensor will continue to take measurements so
        they are available for future reads.

        Prefer using the equivalent :meth:`UltrasonicSensor.distance_inches` property.
        """
        self._ensure_mode(self.MODE_US_DIST_IN)
        return self.value(0) * self._scale('US_DIST_IN')

    @property
    def distance_inches_ping(self):
        """
        Measurement of the distance detected by the sensor,
        in inches.

        The sensor will take a single measurement then stop
        broadcasting.

        If you use this property too frequently (e.g. every
        100msec), the sensor will sometimes lock up and writing
        to the mode attribute will return an error. A delay of
        250msec between each usage seems sufficient to keep the
        sensor from locking up.
        """
        # This mode is special; setting the mode causes the sensor to send out
        # a "ping", but the mode isn't actually changed.
        self.mode = self.MODE_US_SI_IN
        return self.value(0) * self._scale('US_DIST_IN')

    @property
    def distance_inches(self):
        """
        Measurement of the distance detected by the sensor,
        in inches.

        Equivalent to :meth:`UltrasonicSensor.distance_inches_continuous`.
        """
        return self.distance_inches_continuous

    @property
    def other_sensor_present(self):
        """
        Boolean indicating whether another ultrasonic sensor could
        be heard nearby.
        """
        self._ensure_mode(self.MODE_US_LISTEN)
        return bool(self.value(0))


class GyroSensor(Sensor):
    """
    LEGO EV3 gyro sensor.
    """

    SYSTEM_CLASS_NAME = Sensor.SYSTEM_CLASS_NAME
    SYSTEM_DEVICE_NAME_CONVENTION = Sensor.SYSTEM_DEVICE_NAME_CONVENTION

    #: Angle
    MODE_GYRO_ANG = 'GYRO-ANG'

    #: Rotational speed
    MODE_GYRO_RATE = 'GYRO-RATE'

    #: Raw sensor value
    MODE_GYRO_FAS = 'GYRO-FAS'

    #: Angle and rotational speed
    MODE_GYRO_G_A = 'GYRO-G&A'

    #: Calibration ???
    MODE_GYRO_CAL = 'GYRO-CAL'

    # Newer versions of the Gyro sensor also have an additional second axis
    # accessible via the TILT-ANG and TILT-RATE modes that is not usable
    # using the official EV3-G blocks
    MODE_TILT_ANG = 'TILT-ANG'
    MODE_TILT_RATE = 'TILT-RATE'

    MODES = (
        MODE_GYRO_ANG,
        MODE_GYRO_RATE,
        MODE_GYRO_FAS,
        MODE_GYRO_G_A,
        MODE_GYRO_CAL,
        MODE_TILT_ANG,
        MODE_TILT_RATE,
    )

    def __init__(self, address=None, name_pattern=SYSTEM_DEVICE_NAME_CONVENTION, name_exact=False, **kwargs):
        super(GyroSensor, self).__init__(address, name_pattern, name_exact, driver_name='lego-ev3-gyro', **kwargs)
        self._direct = None
        self._init_angle = self.angle

    @property
    def angle(self):
        """
        The number of degrees that the sensor has been rotated
        since it was put into this mode.
        """
        self._ensure_mode(self.MODE_GYRO_ANG)
        return self.value(0)

    @property
    def rate(self):
        """
        The rate at which the sensor is rotating, in degrees/second.
        """
        self._ensure_mode(self.MODE_GYRO_RATE)
        return self.value(0)

    @property
    def angle_and_rate(self):
        """
        Angle (degrees) and Rotational Speed (degrees/second).
        """
        self._ensure_mode(self.MODE_GYRO_G_A)
        return self.value(0), self.value(1)

    @property
    def tilt_angle(self):
        self._ensure_mode(self.MODE_TILT_ANG)
        return self.value(0)

    @property
    def tilt_rate(self):
        self._ensure_mode(self.MODE_TILT_RATE)
        return self.value(0)

    def calibrate(self):
        """
        The robot should be perfectly still when you call this
        """
        current_mode = self.mode
        self._ensure_mode(self.MODE_GYRO_CAL)
        time.sleep(2)
        self._ensure_mode(current_mode)

    def reset(self):
        """Resets the angle to 0.

        Caveats:
            - This function only resets the angle to 0, it does not fix drift.
            - This function only works on EV3, it does not work on BrickPi,
              PiStorms, or with any sensor multiplexors.
        """
        # 17 comes from inspecting the .vix file of the Gyro sensor block in EV3-G
        self._direct = self.set_attr_raw(self._direct, 'direct', b'\x11')
        self._init_angle = self.angle

    def wait_until_angle_changed_by(self, delta, direction_sensitive=False):
        """
        Wait until angle has changed by specified amount.

        If ``direction_sensitive`` is True we will wait until angle has changed
        by ``delta`` and with the correct sign.

        If ``direction_sensitive`` is False (default) we will wait until angle has changed
        by ``delta`` in either direction.
        """
        assert self.mode in (self.MODE_GYRO_G_A, self.MODE_GYRO_ANG,
                             self.MODE_TILT_ANG),\
            'Gyro mode should be MODE_GYRO_ANG, MODE_GYRO_G_A or MODE_TILT_ANG'
        start_angle = self.value(0)

        if direction_sensitive:
            if delta > 0:
                while (self.value(0) - start_angle) < delta:
                    time.sleep(0.01)
            else:
                delta *= -1
                while (start_angle - self.value(0)) < delta:
                    time.sleep(0.01)
        else:
            while abs(start_angle - self.value(0)) < delta:
                time.sleep(0.01)

    def circle_angle(self):
        """
        As the gryo rotates clockwise the angle increases, it will increase
        by 360 for each full rotation. As the gyro rotates counter-clockwise
        the gyro angle will decrease.

        The angles on a circle have the opposite behavior though, they start
        at 0 and increase as you move counter-clockwise around the circle.

        Convert the gyro angle to the angle on a circle. We consider the initial
        position of the gyro to be at 90 degrees on the cirlce.
        """
        current_angle = self.angle
        delta = abs(current_angle - self._init_angle) % 360

        if delta == 0:
            result = 90

        # the gyro has turned clockwise relative to where we started
        elif current_angle > self._init_angle:

            if delta <= 90:
                result = 90 - delta

            elif delta <= 180:
                result = 360 - (delta - 90)

            elif delta <= 270:
                result = 270 - (delta - 180)

            else:
                result = 180 - (delta - 270)

            # This can be chatty (but helpful) so save it for a rainy day
            # log.info("%s moved clockwise %s degrees to %s" % (self, delta, result))

        # the gyro has turned counter-clockwise relative to where we started
        else:
            if delta <= 90:
                result = 90 + delta

            elif delta <= 180:
                result = 180 + (delta - 90)

            elif delta <= 270:
                result = 270 + (delta - 180)

            else:
                result = delta - 270

            # This can be chatty (but helpful) so save it for a rainy day
            # log.info("%s moved counter-clockwise %s degrees to %s" % (self, delta, result))

        return result


class InfraredSensor(Sensor, ButtonBase):
    """
    LEGO EV3 infrared sensor.
    """

    SYSTEM_CLASS_NAME = Sensor.SYSTEM_CLASS_NAME
    SYSTEM_DEVICE_NAME_CONVENTION = Sensor.SYSTEM_DEVICE_NAME_CONVENTION

    #: Proximity
    MODE_IR_PROX = 'IR-PROX'

    #: IR Seeker
    MODE_IR_SEEK = 'IR-SEEK'

    #: IR Remote Control
    MODE_IR_REMOTE = 'IR-REMOTE'

    #: IR Remote Control. State of the buttons is coded in binary
    MODE_IR_REM_A = 'IR-REM-A'

    #: Calibration ???
    MODE_IR_CAL = 'IR-CAL'

    MODES = (MODE_IR_PROX, MODE_IR_SEEK, MODE_IR_REMOTE, MODE_IR_REM_A, MODE_IR_CAL)

    # The following are all of the various combinations of button presses for
    # the remote control.  The key/index is the number that will be written in
    # the attribute file to indicate what combination of buttons are currently
    # pressed.
    _BUTTON_VALUES = {
        0: [],
        1: ['top_left'],
        2: ['bottom_left'],
        3: ['top_right'],
        4: ['bottom_right'],
        5: ['top_left', 'top_right'],
        6: ['top_left', 'bottom_right'],
        7: ['bottom_left', 'top_right'],
        8: ['bottom_left', 'bottom_right'],
        9: ['beacon'],
        10: ['top_left', 'bottom_left'],
        11: ['top_right', 'bottom_right']
    }

    _BUTTONS = ('top_left', 'bottom_left', 'top_right', 'bottom_right', 'beacon')

    # Button codes for doing rapid check of remote status
    NO_BUTTON = 0
    TOP_LEFT = 1
    BOTTOM_LEFT = 2
    TOP_RIGHT = 3
    BOTTOM_RIGHT = 4
    TOP_LEFT_TOP_RIGHT = 5
    TOP_LEFT_BOTTOM_RIGHT = 6
    BOTTOM_LEFT_TOP_RIGHT = 7
    BOTTOM_LEFT_BOTTOM_RIGHT = 8
    BEACON = 9
    TOP_LEFT_BOTTOM_LEFT = 10
    TOP_RIGHT_BOTTOM_RIGHT = 11

    #: Handler for top-left button events on channel 1. See :meth:`InfraredSensor.process`.
    on_channel1_top_left = None
    #: Handler for bottom-left button events on channel 1. See :meth:`InfraredSensor.process`.
    on_channel1_bottom_left = None
    #: Handler for top-right button events on channel 1. See :meth:`InfraredSensor.process`.
    on_channel1_top_right = None
    #: Handler for bottom-right button events on channel 1. See :meth:`InfraredSensor.process`.
    on_channel1_bottom_right = None
    #: Handler for beacon button events on channel 1. See :meth:`InfraredSensor.process`.
    on_channel1_beacon = None

    #: Handler for top-left button events on channel 2. See :meth:`InfraredSensor.process`.
    on_channel2_top_left = None
    #: Handler for bottom-left button events on channel 2. See :meth:`InfraredSensor.process`.
    on_channel2_bottom_left = None
    #: Handler for top-right button events on channel 2. See :meth:`InfraredSensor.process`.
    on_channel2_top_right = None
    #: Handler for bottom-right button events on channel 2. See :meth:`InfraredSensor.process`.
    on_channel2_bottom_right = None
    #: Handler for beacon button events on channel 2. See :meth:`InfraredSensor.process`.
    on_channel2_beacon = None

    #: Handler for top-left button events on channel 3. See :meth:`InfraredSensor.process`.
    on_channel3_top_left = None
    #: Handler for bottom-left button events on channel 3. See :meth:`InfraredSensor.process`.
    on_channel3_bottom_left = None
    #: Handler for top-right button events on channel 3. See :meth:`InfraredSensor.process`.
    on_channel3_top_right = None
    #: Handler for bottom-right button events on channel 3. See :meth:`InfraredSensor.process`.
    on_channel3_bottom_right = None
    #: Handler for beacon button events on channel 3. See :meth:`InfraredSensor.process`.
    on_channel3_beacon = None

    #: Handler for top-left button events on channel 4. See :meth:`InfraredSensor.process`.
    on_channel4_top_left = None
    #: Handler for bottom-left button events on channel 4. See :meth:`InfraredSensor.process`.
    on_channel4_bottom_left = None
    #: Handler for top-right button events on channel 4. See :meth:`InfraredSensor.process`.
    on_channel4_top_right = None
    #: Handler for bottom-right button events on channel 4. See :meth:`InfraredSensor.process`.
    on_channel4_bottom_right = None
    #: Handler for beacon button events on channel 4. See :meth:`InfraredSensor.process`.
    on_channel4_beacon = None

    def __init__(self, address=None, name_pattern=SYSTEM_DEVICE_NAME_CONVENTION, name_exact=False, **kwargs):
        super(InfraredSensor, self).__init__(address, name_pattern, name_exact, driver_name='lego-ev3-ir', **kwargs)

    def _normalize_channel(self, channel):
        assert channel >= 1 and channel <= 4, "channel is %s, it must be 1, 2, 3, or 4" % channel
        channel = max(1, min(4, channel)) - 1
        return channel

    @property
    def proximity(self):
        """
        An estimate of the distance between the sensor and objects in front of
        it, as a percentage. 100% is approximately 70cm/27in.
        """
        self._ensure_mode(self.MODE_IR_PROX)
        return self.value(0)

    def heading(self, channel=1):
        """
        Returns heading (-25, 25) to the beacon on the given channel.
        """
        self._ensure_mode(self.MODE_IR_SEEK)
        channel = self._normalize_channel(channel)
        return self.value(channel * 2)

    def distance(self, channel=1):
        """
        Returns distance (0, 100) to the beacon on the given channel.
        Returns None when beacon is not found.
        """
        self._ensure_mode(self.MODE_IR_SEEK)
        channel = self._normalize_channel(channel)
        ret_value = self.value((channel * 2) + 1)

        # The value will be -128 if no beacon is found, return None instead
        return None if ret_value == -128 else ret_value

    def heading_and_distance(self, channel=1):
        """
        Returns heading and distance to the beacon on the given channel as a
        tuple.
        """
        return (self.heading(channel), self.distance(channel))

    def top_left(self, channel=1):
        """
        Checks if ``top_left`` button is pressed.
        """
        return 'top_left' in self.buttons_pressed(channel)

    def bottom_left(self, channel=1):
        """
        Checks if ``bottom_left`` button is pressed.
        """
        return 'bottom_left' in self.buttons_pressed(channel)

    def top_right(self, channel=1):
        """
        Checks if ``top_right`` button is pressed.
        """
        return 'top_right' in self.buttons_pressed(channel)

    def bottom_right(self, channel=1):
        """
        Checks if ``bottom_right`` button is pressed.
        """
        return 'bottom_right' in self.buttons_pressed(channel)

    def beacon(self, channel=1):
        """
        Checks if ``beacon`` button is pressed.
        """
        return 'beacon' in self.buttons_pressed(channel)

    def buttons_pressed(self, channel=1):
        """
        Returns list of currently pressed buttons.

        Note that the sensor can only identify up to two buttons pressed at once.
        """
        self._ensure_mode(self.MODE_IR_REMOTE)
        channel = self._normalize_channel(channel)
        return self._BUTTON_VALUES.get(self.value(channel), [])

    def process(self):
        """
        Check for currenly pressed buttons. If the new state differs from the
        old state, call the appropriate button event handlers.

        To use the on_channel1_top_left, etc handlers your program would do something like:

        .. code:: python

            def top_left_channel_1_action(state):
                print("top left on channel 1: %s" % state)

            def bottom_right_channel_4_action(state):
                print("bottom right on channel 4: %s" % state)

            ir = InfraredSensor()
            ir.on_channel1_top_left = top_left_channel_1_action
            ir.on_channel4_bottom_right = bottom_right_channel_4_action

            while True:
                ir.process()
                time.sleep(0.01)

        """
        new_state = []
        state_diff = []

        for channel in range(1, 5):

            for button in self.buttons_pressed(channel):
                new_state.append((button, channel))

                # Key was not pressed before but now is pressed
                if (button, channel) not in self._state:
                    state_diff.append((button, channel))

            # Key was pressed but is no longer pressed
            for button in self._BUTTONS:
                if (button, channel) not in new_state and (button, channel) in self._state:
                    state_diff.append((button, channel))

        self._state = new_state

        for (button, channel) in state_diff:
            handler = getattr(self, 'on_channel' + str(channel) + '_' + button)

            if handler is not None:
                handler((button, channel) in new_state)

        if self.on_change is not None and state_diff:
            self.on_change([(button, channel, button in new_state) for (button, channel) in state_diff])


class SoundSensor(Sensor):
    """
    LEGO NXT Sound Sensor
    """

    SYSTEM_CLASS_NAME = Sensor.SYSTEM_CLASS_NAME
    SYSTEM_DEVICE_NAME_CONVENTION = Sensor.SYSTEM_DEVICE_NAME_CONVENTION

    #: Sound pressure level. Flat weighting
    MODE_DB = 'DB'

    #: Sound pressure level. A weighting
    MODE_DBA = 'DBA'

    MODES = (
        MODE_DB,
        MODE_DBA,
    )

    def __init__(self, address=None, name_pattern=SYSTEM_DEVICE_NAME_CONVENTION, name_exact=False, **kwargs):
        super(SoundSensor, self).__init__(address, name_pattern, name_exact, driver_name='lego-nxt-sound', **kwargs)

    @property
    def sound_pressure(self):
        """
        A measurement of the measured sound pressure level, as a
        percent. Uses a flat weighting.
        """
        self._ensure_mode(self.MODE_DB)
        return self.value(0) * self._scale('DB')

    @property
    def sound_pressure_low(self):
        """
        A measurement of the measured sound pressure level, as a
        percent. Uses A-weighting, which focuses on levels up to 55 dB.
        """
        self._ensure_mode(self.MODE_DBA)
        return self.value(0) * self._scale('DBA')


class LightSensor(Sensor):
    """
    LEGO NXT Light Sensor
    """

    SYSTEM_CLASS_NAME = Sensor.SYSTEM_CLASS_NAME
    SYSTEM_DEVICE_NAME_CONVENTION = Sensor.SYSTEM_DEVICE_NAME_CONVENTION

    #: Reflected light. LED on
    MODE_REFLECT = 'REFLECT'

    #: Ambient light. LED off
    MODE_AMBIENT = 'AMBIENT'

    MODES = (
        MODE_REFLECT,
        MODE_AMBIENT,
    )

    def __init__(self, address=None, name_pattern=SYSTEM_DEVICE_NAME_CONVENTION, name_exact=False, **kwargs):
        super(LightSensor, self).__init__(address, name_pattern, name_exact, driver_name='lego-nxt-light', **kwargs)

    @property
    def reflected_light_intensity(self):
        """
        A measurement of the reflected light intensity, as a percentage.
        """
        self._ensure_mode(self.MODE_REFLECT)
        return self.value(0) * self._scale('REFLECT')

    @property
    def ambient_light_intensity(self):
        """
        A measurement of the ambient light intensity, as a percentage.
        """
        self._ensure_mode(self.MODE_AMBIENT)
        return self.value(0) * self._scale('AMBIENT')



import ast
import xml.etree.ElementTree as ET
import mindpile.Utility.utility as utility
import mindpile.Parser as Parser
from mindpile.Mapping.utils import generateSetupAST, startCodeGen

__version__ = "0.0.1"

def to_python(ev3p_path, ev3sim_support=False):
    startCodeGen()
    tree = ET.parse(ev3p_path)
    root = tree.getroot()

    for elem in root.iter(utility.addNameSpace("BlockDiagram")):
        root = elem

    codeBlock = Parser.tagHandler.translateElementToIRForm(root)
    codeBlock.sortInternalFlow()
    tree = ast.parse('')
    main = codeBlock.toAST(ctx={"ev3sim_support": ev3sim_support})
    tree.body += generateSetupAST().body
    tree.body += main
    tree = ast.fix_missing_locations(tree)
    return ast.unparse(tree)

def from_ev3(ev3_path, ev3sim_support=False):
    import os
    import tempfile
    from zipfile import ZipFile
    # Some preliminary checks
    if not ev3_path.endswith(".ev3"):
        raise ValueError("The mindstorms file to extract should have the `.ev3` file ending.")
    # First -> Unzip the .ev3 file into a temporary location. Extract the ev3p file.
    with tempfile.TemporaryDirectory() as tmpdir:
        with ZipFile(ev3_path, "r") as zipObj:
            zipObj.extractall(tmpdir)
        # Second -> Generate the python code.
        python_source = to_python(os.path.join(tmpdir, "Program.ev3p"), ev3sim_support=ev3sim_support)
        return python_source


from mindpile.Mapping.utils import MethodCall

@MethodCall(target="Comparison_Equal.vix", x=float, y=float, Result=bool)
def equal():
    return '''
        Result = x == y
    '''

@MethodCall(target="Comparison_NotEqual.vix", x=float, y=float, Result=bool)
def notEqual():
    return '''
        Result = x != y
    '''

@MethodCall(target="Comparison_Greater.vix", x=float, y=float, Result=bool)
def greater():
    return '''
        Result = x > y
    '''

@MethodCall(target="Comparison_GreaterEqual.vix", x=float, y=float, Result=bool)
def greaterEqual():
    return '''
        Result = x >= y
    '''

@MethodCall(target="Comparison_Less.vix", x=float, y=float, Result=bool)
def less():
    return '''
        Result = x < y
    '''

@MethodCall(target="Comparison_LessEqual.vix", x=float, y=float, Result=bool)
def lessEqual():
    return '''
        Result = x <= y
    '''



from mindpile.Mapping.utils import MethodCall

@MethodCall(target="Round_Nearest.vix", Input=float, OutputResult=float)
def roundNearest():
    return '''
        OutputResult = round(Input)
    '''

@MethodCall(target="Round_Up.vix", Input=float, OutputResult=float)
def roundUp():
    return '''
        OutputResult = math.ceil(Input)
    '''

@MethodCall(target="Round_Down.vix", Input=float, OutputResult=float)
def roundDown():
    return '''
        OutputResult = math.floor(Input)
    '''

@MethodCall(target="Round_Truncate.vix", Input=float, NumberofDecimals=int, OutputResult=float)
def roundTruncate():
    return '''
        OutputResult = round(Input, NumberofDecimals)
    '''



# -----------------------------------------------------------------------------
# Copyright (c) 2015 Ralph Hempel <rhempel@hempeldesigngroup.com>
# Copyright (c) 2015 Anton Vanhoucke <antonvh@gmail.com>
# Copyright (c) 2015 Denis Demidov <dennis.demidov@gmail.com>
# Copyright (c) 2015 Eric Pascual <eric@pobot.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# -----------------------------------------------------------------------------

import sys
import os
import re
from time import sleep
from ev3dev2 import is_micropython

if sys.version_info < (3, 4):
    raise SystemError('Must be using Python 3.4 or higher')

if not is_micropython():
    import shlex
    from subprocess import Popen, PIPE


def _make_scales(notes):
    """ Utility function used by Sound class for building the note frequencies table """
    res = dict()
    for note, freq in notes:
        freq = round(freq)
        for n in note.split('/'):
            res[n.upper()] = freq
    return res


def get_command_processes(command):
    """
    :param string command: a string of command(s) to run that may include pipes
    :return: a list of Popen objects
    """

    # We must split command into sub-commands to support pipes
    if "|" in command:
        command_parts = command.split("|")
    else:
        command_parts = [command]

    processes = []

    for command_part in command_parts:
        if processes:
            processes.append(Popen(shlex.split(command_part), stdin=processes[-1].stdout, stdout=PIPE, stderr=PIPE))
        else:
            processes.append(Popen(shlex.split(command_part), stdin=None, stdout=PIPE, stderr=PIPE))

    return processes


class Sound(object):
    """
    Support beep, play wav files, or convert text to speech.

    Examples::

        from ev3dev2.sound import Sound

        spkr = Sound()

        # Play 'bark.wav':
        spkr.play_file('bark.wav')

        # Introduce yourself:
        spkr.speak('Hello, I am Robot')

        # Play a small song
        spkr.play_song((
            ('D4', 'e3'),
            ('D4', 'e3'),
            ('D4', 'e3'),
            ('G4', 'h'),
            ('D5', 'h')
        ))

    In order to mimic EV3-G API parameters, durations used in methods
    exposed as EV3-G blocks for sound related operations are expressed
    as a float number of seconds.
    """

    channel = None

    # play_types
    PLAY_WAIT_FOR_COMPLETE = 0  #: Play the sound and block until it is complete
    PLAY_NO_WAIT_FOR_COMPLETE = 1  #: Start playing the sound but return immediately
    PLAY_LOOP = 2  #: Never return; start the sound immediately after it completes, until the program is killed

    PLAY_TYPES = (PLAY_WAIT_FOR_COMPLETE, PLAY_NO_WAIT_FOR_COMPLETE, PLAY_LOOP)

    def _validate_play_type(self, play_type):
        assert play_type in self.PLAY_TYPES, \
            "Invalid play_type %s, must be one of %s" % (play_type, ','.join(str(t) for t in self.PLAY_TYPES))

    def _audio_command(self, command, play_type):
        if is_micropython():

            if play_type == Sound.PLAY_WAIT_FOR_COMPLETE:
                os.system(command)

            elif play_type == Sound.PLAY_NO_WAIT_FOR_COMPLETE:
                os.system('{} &'.format(command))

            elif play_type == Sound.PLAY_LOOP:
                while True:
                    os.system(command)

            else:
                raise Exception("invalid play_type " % play_type)

            return None

        else:
            with open(os.devnull, 'w'):

                if play_type == Sound.PLAY_WAIT_FOR_COMPLETE:
                    processes = get_command_processes(command)
                    processes[-1].communicate()
                    processes[-1].wait()
                    return None

                elif play_type == Sound.PLAY_NO_WAIT_FOR_COMPLETE:
                    processes = get_command_processes(command)
                    return processes[-1]

                elif play_type == Sound.PLAY_LOOP:
                    while True:
                        processes = get_command_processes(command)
                        processes[-1].communicate()
                        processes[-1].wait()

                else:
                    raise Exception("invalid play_type " % play_type)

    def beep(self, args='', play_type=PLAY_WAIT_FOR_COMPLETE):
        """
        Call beep command with the provided arguments (if any).
        See `beep man page`_ and google `linux beep music`_ for inspiration.

        :param string args: Any additional arguments to be passed to ``beep`` (see the `beep man page`_ for details)

        :param play_type: The behavior of ``beep`` once playback has been initiated
        :type play_type: ``Sound.PLAY_WAIT_FOR_COMPLETE`` or  ``Sound.PLAY_NO_WAIT_FOR_COMPLETE``

        :return: When python3 is used and ``Sound.PLAY_NO_WAIT_FOR_COMPLETE`` is specified, returns the
            spawn subprocess from ``subprocess.Popen``; ``None`` otherwise

        .. _`beep man page`: https://linux.die.net/man/1/beep
        .. _`linux beep music`: https://www.google.com/search?q=linux+beep+music
        """
        return self._audio_command("/usr/bin/beep %s" % args, play_type)

    def tone(self, *args, play_type=PLAY_WAIT_FOR_COMPLETE):
        """
        .. rubric:: tone(tone_sequence)

        Play tone sequence.

        Here is a cheerful example::

            my_sound = Sound()
            my_sound.tone([
                (392, 350, 100), (392, 350, 100), (392, 350, 100), (311.1, 250, 100),
                (466.2, 25, 100), (392, 350, 100), (311.1, 250, 100), (466.2, 25, 100),
                (392, 700, 100), (587.32, 350, 100), (587.32, 350, 100),
                (587.32, 350, 100), (622.26, 250, 100), (466.2, 25, 100),
                (369.99, 350, 100), (311.1, 250, 100), (466.2, 25, 100), (392, 700, 100),
                (784, 350, 100), (392, 250, 100), (392, 25, 100), (784, 350, 100),
                (739.98, 250, 100), (698.46, 25, 100), (659.26, 25, 100),
                (622.26, 25, 100), (659.26, 50, 400), (415.3, 25, 200), (554.36, 350, 100),
                (523.25, 250, 100), (493.88, 25, 100), (466.16, 25, 100), (440, 25, 100),
                (466.16, 50, 400), (311.13, 25, 200), (369.99, 350, 100),
                (311.13, 250, 100), (392, 25, 100), (466.16, 350, 100), (392, 250, 100),
                (466.16, 25, 100), (587.32, 700, 100), (784, 350, 100), (392, 250, 100),
                (392, 25, 100), (784, 350, 100), (739.98, 250, 100), (698.46, 25, 100),
                (659.26, 25, 100), (622.26, 25, 100), (659.26, 50, 400), (415.3, 25, 200),
                (554.36, 350, 100), (523.25, 250, 100), (493.88, 25, 100),
                (466.16, 25, 100), (440, 25, 100), (466.16, 50, 400), (311.13, 25, 200),
                (392, 350, 100), (311.13, 250, 100), (466.16, 25, 100),
                (392.00, 300, 150), (311.13, 250, 100), (466.16, 25, 100), (392, 700)
                ])

        Have also a look at :py:meth:`play_song` for a more musician-friendly way of doing, which uses
        the conventional notation for notes and durations.

        :param list[tuple(float,float,float)] tone_sequence: The sequence of tones to play. The first
            number of each tuple is frequency in Hz, the second is duration in milliseconds, and the
            third is delay in milliseconds between this and the next tone in the sequence.

        :param play_type: The behavior of ``tone`` once playback has been initiated
        :type play_type: ``Sound.PLAY_WAIT_FOR_COMPLETE`` or ``Sound.PLAY_NO_WAIT_FOR_COMPLETE``

        :return: When python3 is used and ``Sound.PLAY_NO_WAIT_FOR_COMPLETE`` is specified, returns the
            spawn subprocess from ``subprocess.Popen``; ``None`` otherwise

        .. rubric:: tone(frequency, duration)

        Play single tone of given frequency and duration.

        :param float frequency: The frequency of the tone in Hz
        :param float duration: The duration of the tone in milliseconds

        :param play_type: The behavior of ``tone`` once playback has been initiated
        :type play_type: ``Sound.PLAY_WAIT_FOR_COMPLETE`` or ``Sound.PLAY_NO_WAIT_FOR_COMPLETE``

        :return: When python3 is used and ``Sound.PLAY_NO_WAIT_FOR_COMPLETE`` is specified, returns the
            spawn subprocess from ``subprocess.Popen``; ``None`` otherwise
        """
        def play_tone_sequence(tone_sequence):
            def beep_args(frequency=None, duration=None, delay=None):
                args = ''
                if frequency is not None:
                    args += '-f %s ' % frequency
                if duration is not None:
                    args += '-l %s ' % duration
                if delay is not None:
                    args += '-D %s ' % delay

                return args

            return self.beep(' -n '.join([beep_args(*t) for t in tone_sequence]), play_type=play_type)

        if len(args) == 1:
            return play_tone_sequence(args[0])
        elif len(args) == 2:
            return play_tone_sequence([(args[0], args[1])])
        else:
            raise Exception("Unsupported number of parameters in Sound.tone(): expected 1 or 2, got " + str(len(args)))

    def play_tone(self, frequency, duration, delay=0.0, volume=100, play_type=PLAY_WAIT_FOR_COMPLETE):
        """ Play a single tone, specified by its frequency, duration, volume and final delay.

        :param int frequency: the tone frequency, in Hertz
        :param float duration: Tone duration, in seconds
        :param float delay: Delay after tone, in seconds (can be useful when chaining calls to ``play_tone``)
        :param int volume: The play volume, in percent of maximum volume

        :param play_type: The behavior of ``play_tone`` once playback has been initiated
        :type play_type: ``Sound.PLAY_WAIT_FOR_COMPLETE``, ``Sound.PLAY_NO_WAIT_FOR_COMPLETE`` or ``Sound.PLAY_LOOP``

        :return: When python3 is used and ``Sound.PLAY_NO_WAIT_FOR_COMPLETE`` is specified, returns
            the PID of the underlying beep command; ``None`` otherwise

        :raises ValueError: if invalid parameter
        """
        self._validate_play_type(play_type)

        if duration <= 0:
            raise ValueError('invalid duration (%s)' % duration)
        if delay < 0:
            raise ValueError('invalid delay (%s)' % delay)
        if not 0 < volume <= 100:
            raise ValueError('invalid volume (%s)' % volume)

        self.set_volume(volume)

        duration_ms = int(duration * 1000)
        delay_ms = int(delay * 1000)

        self.tone([(frequency, duration_ms, delay_ms)], play_type=play_type)

    def play_note(self, note, duration, volume=100, play_type=PLAY_WAIT_FOR_COMPLETE):
        """ Plays a note, given by its name as defined in ``_NOTE_FREQUENCIES``.

        :param string note: The note symbol with its octave number
        :param float duration: Tone duration, in seconds
        :param int volume: The play volume, in percent of maximum volume

        :param play_type: The behavior of ``play_note`` once playback has been initiated
        :type play_type: ``Sound.PLAY_WAIT_FOR_COMPLETE``, ``Sound.PLAY_NO_WAIT_FOR_COMPLETE`` or ``Sound.PLAY_LOOP``

        :return: When python3 is used and ``Sound.PLAY_NO_WAIT_FOR_COMPLETE`` is specified, returns
            the PID of the underlying beep command; ``None`` otherwise

        :raises ValueError: is invalid parameter (note, duration,...)
        """
        self._validate_play_type(play_type)
        try:
            freq = self._NOTE_FREQUENCIES.get(note.upper(), self._NOTE_FREQUENCIES[note])
        except KeyError:
            raise ValueError('invalid note (%s)' % note)

        if duration <= 0:
            raise ValueError('invalid duration (%s)' % duration)
        if not 0 < volume <= 100:
            raise ValueError('invalid volume (%s)' % volume)

        return self.play_tone(freq, duration=duration, volume=volume, play_type=play_type)

    def play_file(self, wav_file, volume=100, play_type=PLAY_WAIT_FOR_COMPLETE):
        """ Play a sound file (wav format) at a given volume.

        :param string wav_file: The sound file path
        :param int volume: The play volume, in percent of maximum volume

        :param play_type: The behavior of ``play_file`` once playback has been initiated
        :type play_type: ``Sound.PLAY_WAIT_FOR_COMPLETE``, ``Sound.PLAY_NO_WAIT_FOR_COMPLETE`` or ``Sound.PLAY_LOOP``

        :return: When python3 is used and ``Sound.PLAY_NO_WAIT_FOR_COMPLETE`` is specified, returns the
            spawn subprocess from ``subprocess.Popen``; ``None`` otherwise
        """
        if not 0 < volume <= 100:
            raise ValueError('invalid volume (%s)' % volume)

        if not wav_file.endswith(".wav"):
            raise ValueError('invalid sound file (%s), only .wav files are supported' % wav_file)

        if not os.path.exists(wav_file):
            raise ValueError("%s does not exist" % wav_file)

        self._validate_play_type(play_type)
        self.set_volume(volume)
        return self._audio_command('/usr/bin/aplay -q "%s"' % wav_file, play_type)

    def speak(self, text, espeak_opts='-a 200 -s 130', volume=100, play_type=PLAY_WAIT_FOR_COMPLETE):
        """ Speak the given text aloud.

        Uses the ``espeak`` external command.

        :param string text: The text to speak
        :param string espeak_opts: ``espeak`` command options (advanced usage)
        :param int volume: The play volume, in percent of maximum volume

        :param play_type: The behavior of ``speak`` once playback has been initiated
        :type play_type: ``Sound.PLAY_WAIT_FOR_COMPLETE``, ``Sound.PLAY_NO_WAIT_FOR_COMPLETE`` or ``Sound.PLAY_LOOP``

        :return: When python3 is used and ``Sound.PLAY_NO_WAIT_FOR_COMPLETE`` is specified, returns the
            spawn subprocess from ``subprocess.Popen``; ``None`` otherwise
        """
        self._validate_play_type(play_type)
        self.set_volume(volume)
        cmd = "/usr/bin/espeak --stdout %s '%s' | /usr/bin/aplay -q" % (espeak_opts, text)
        return self._audio_command(cmd, play_type)

    def _get_channel(self):
        """
        :return: The detected sound channel
        :rtype: string
        """
        if self.channel is None:
            # Get default channel as the first one that pops up in
            # 'amixer scontrols' output, which contains strings in the
            # following format:
            #
            #     Simple mixer control 'Master',0
            #     Simple mixer control 'Capture',0
            out = os.popen('/usr/bin/amixer scontrols').read()
            m = re.search(r"'([^']+)'", out)
            if m:
                self.channel = m.group(1)
            else:
                self.channel = 'Playback'

        return self.channel

    def set_volume(self, pct, channel=None):
        """
        Sets the sound volume to the given percentage [0-100] by calling
        ``amixer -q set <channel> <pct>%``.
        If the channel is not specified, it tries to determine the default one
        by running ``amixer scontrols``. If that fails as well, it uses the
        ``Playback`` channel, as that is the only channel on the EV3.
        """

        if channel is None:
            channel = self._get_channel()

        os.system('/usr/bin/amixer -q set {0} {1:d}%'.format(channel, pct))

    def get_volume(self, channel=None):
        """
        Gets the current sound volume by parsing the output of
        ``amixer get <channel>``.
        If the channel is not specified, it tries to determine the default one
        by running ``amixer scontrols``. If that fails as well, it uses the
        ``Playback`` channel, as that is the only channel on the EV3.
        """

        if channel is None:
            channel = self._get_channel()

        out = os.popen(['/usr/bin/amixer', 'get', channel]).read()
        m = re.search(r'\[(\d+)%\]', out)
        if m:
            return int(m.group(1))
        else:
            raise Exception('Failed to parse output of ``amixer get {}``'.format(channel))

    def play_song(self, song, tempo=120, delay=0.05):
        """ Plays a song provided as a list of tuples containing the note name and its
        value using music conventional notation instead of numerical values for frequency
        and duration.

        It supports symbolic notes (e.g. ``A4``, ``D#3``, ``Gb5``) and durations (e.g. ``q``, ``h``).
        You can also specify rests by using ``R`` instead of note pitch.

        For an exhaustive list of accepted note symbols and values, have a look at the ``_NOTE_FREQUENCIES``
        and ``_NOTE_VALUES`` private dictionaries in the source code.

        The value can be suffixed by modifiers:

        - a *divider* introduced by a ``/`` to obtain triplets for instance
          (e.g. ``q/3`` for a triplet of eight note)
        - a *multiplier* introduced by ``*`` (e.g. ``*1.5`` is a dotted note).

        Shortcuts exist for common modifiers:

        - ``3`` produces a triplet member note. For instance ``e3`` gives a triplet of eight notes,
          i.e. 3 eight notes in the duration of a single quarter. You must ensure that 3 triplets
          notes are defined in sequence to match the count, otherwise the result will not be the
          expected one.
        - ``.`` produces a dotted note, i.e. which duration is one and a half the base one. Double dots
          are not currently supported.

        Example::

            >>> # A long time ago in a galaxy far,
            >>> # far away...
            >>> from ev3dev2.sound import Sound
            >>> spkr = Sound()
            >>> spkr.play_song((
            >>>     ('D4', 'e3'),      # intro anacrouse
            >>>     ('D4', 'e3'),
            >>>     ('D4', 'e3'),
            >>>     ('G4', 'h'),       # meas 1
            >>>     ('D5', 'h'),
            >>>     ('C5', 'e3'),      # meas 2
            >>>     ('B4', 'e3'),
            >>>     ('A4', 'e3'),
            >>>     ('G5', 'h'),
            >>>     ('D5', 'q'),
            >>>     ('C5', 'e3'),      # meas 3
            >>>     ('B4', 'e3'),
            >>>     ('A4', 'e3'),
            >>>     ('G5', 'h'),
            >>>     ('D5', 'q'),
            >>>     ('C5', 'e3'),      # meas 4
            >>>     ('B4', 'e3'),
            >>>     ('C5', 'e3'),
            >>>     ('A4', 'h.'),
            >>> ))

        .. important::

            Only 4/4 signature songs are supported with respect to note durations.

        :param iterable[tuple(string,string)] song: the song
        :param int tempo: the song tempo, given in quarters per minute
        :param float delay: delay between notes (in seconds)

        :return: When python3 is used the spawn subprocess from ``subprocess.Popen`` is returned; ``None`` otherwise

        :raises ValueError: if invalid note in song or invalid play parameters
        """
        if tempo <= 0:
            raise ValueError('invalid tempo (%s)' % tempo)
        if delay < 0:
            raise ValueError('invalid delay (%s)' % delay)

        delay_ms = int(delay * 1000)
        meas_duration_ms = 60000 / tempo * 4  # we only support 4/4 bars, hence "* 4"

        for (note, value) in song:
            value = value.lower()

            if '/' in value:
                base, factor = value.split('/')
                factor = float(factor)

            elif '*' in value:
                base, factor = value.split('*')
                factor = float(factor)

            elif value.endswith('.'):
                base = value[:-1]
                factor = 1.5

            elif value.endswith('3'):
                base = value[:-1]
                factor = float(2 / 3)

            else:
                base = value
                factor = 1.0

            try:
                duration_ms = meas_duration_ms * self._NOTE_VALUES[base] * factor
            except KeyError:
                raise ValueError('invalid note (%s)' % base)

            if note == "R":
                sleep(duration_ms / 1000 + delay)
            else:
                freq = self._NOTE_FREQUENCIES[note.upper()]
                self.beep('-f %d -l %d -D %d' % (freq, duration_ms, delay_ms))

    #: Note frequencies.
    #:
    #: This dictionary gives the rounded frequency of a note specified by its
    #: standard US abbreviation and its octave number (e.g. ``C3``).
    #: Alterations use the ``#`` and ``b`` symbols, respectively for
    #: *sharp* and *flat*, between the note code and the octave number (e.g. ``D#4``, ``Gb5``).
    _NOTE_FREQUENCIES = _make_scales((
        ('C0', 16.35),
        ('C#0/Db0', 17.32),
        ('D0', 18.35),
        ('D#0/Eb0', 19.45),  # expanded in one entry per symbol by _make_scales
        ('E0', 20.60),
        ('F0', 21.83),
        ('F#0/Gb0', 23.12),
        ('G0', 24.50),
        ('G#0/Ab0', 25.96),
        ('A0', 27.50),
        ('A#0/Bb0', 29.14),
        ('B0', 30.87),
        ('C1', 32.70),
        ('C#1/Db1', 34.65),
        ('D1', 36.71),
        ('D#1/Eb1', 38.89),
        ('E1', 41.20),
        ('F1', 43.65),
        ('F#1/Gb1', 46.25),
        ('G1', 49.00),
        ('G#1/Ab1', 51.91),
        ('A1', 55.00),
        ('A#1/Bb1', 58.27),
        ('B1', 61.74),
        ('C2', 65.41),
        ('C#2/Db2', 69.30),
        ('D2', 73.42),
        ('D#2/Eb2', 77.78),
        ('E2', 82.41),
        ('F2', 87.31),
        ('F#2/Gb2', 92.50),
        ('G2', 98.00),
        ('G#2/Ab2', 103.83),
        ('A2', 110.00),
        ('A#2/Bb2', 116.54),
        ('B2', 123.47),
        ('C3', 130.81),
        ('C#3/Db3', 138.59),
        ('D3', 146.83),
        ('D#3/Eb3', 155.56),
        ('E3', 164.81),
        ('F3', 174.61),
        ('F#3/Gb3', 185.00),
        ('G3', 196.00),
        ('G#3/Ab3', 207.65),
        ('A3', 220.00),
        ('A#3/Bb3', 233.08),
        ('B3', 246.94),
        ('C4', 261.63),
        ('C#4/Db4', 277.18),
        ('D4', 293.66),
        ('D#4/Eb4', 311.13),
        ('E4', 329.63),
        ('F4', 349.23),
        ('F#4/Gb4', 369.99),
        ('G4', 392.00),
        ('G#4/Ab4', 415.30),
        ('A4', 440.00),
        ('A#4/Bb4', 466.16),
        ('B4', 493.88),
        ('C5', 523.25),
        ('C#5/Db5', 554.37),
        ('D5', 587.33),
        ('D#5/Eb5', 622.25),
        ('E5', 659.25),
        ('F5', 698.46),
        ('F#5/Gb5', 739.99),
        ('G5', 783.99),
        ('G#5/Ab5', 830.61),
        ('A5', 880.00),
        ('A#5/Bb5', 932.33),
        ('B5', 987.77),
        ('C6', 1046.50),
        ('C#6/Db6', 1108.73),
        ('D6', 1174.66),
        ('D#6/Eb6', 1244.51),
        ('E6', 1318.51),
        ('F6', 1396.91),
        ('F#6/Gb6', 1479.98),
        ('G6', 1567.98),
        ('G#6/Ab6', 1661.22),
        ('A6', 1760.00),
        ('A#6/Bb6', 1864.66),
        ('B6', 1975.53),
        ('C7', 2093.00),
        ('C#7/Db7', 2217.46),
        ('D7', 2349.32),
        ('D#7/Eb7', 2489.02),
        ('E7', 2637.02),
        ('F7', 2793.83),
        ('F#7/Gb7', 2959.96),
        ('G7', 3135.96),
        ('G#7/Ab7', 3322.44),
        ('A7', 3520.00),
        ('A#7/Bb7', 3729.31),
        ('B7', 3951.07),
        ('C8', 4186.01),
        ('C#8/Db8', 4434.92),
        ('D8', 4698.63),
        ('D#8/Eb8', 4978.03),
        ('E8', 5274.04),
        ('F8', 5587.65),
        ('F#8/Gb8', 5919.91),
        ('G8', 6271.93),
        ('G#8/Ab8', 6644.88),
        ('A8', 7040.00),
        ('A#8/Bb8', 7458.62),
        ('B8', 7902.13)))

    #: Common note values.
    #:
    #: See https://en.wikipedia.org/wiki/Note_value
    #:
    #: This dictionary provides the multiplier to be applied to de whole note duration
    #: to obtain subdivisions, given the corresponding symbolic identifier:
    #:
    #:  = ===============================
    #:  w whole note (UK: semibreve)
    #:  h half note (UK: minim)
    #:  q quarter note (UK: crotchet)
    #:  e eight note (UK: quaver)
    #:  s sixteenth note (UK: semiquaver)
    #:  = ===============================
    #:
    #:
    #: Triplets can be obtained by dividing the corresponding reference by 3.
    #: For instance, the note value of a eight triplet will be ``NOTE_VALUE['e'] / 3``.
    #: It is simpler however to user the ``3`` modifier of notes, as supported by the
    #: :py:meth:`Sound.play_song` method.
    _NOTE_VALUES = {
        'w': 1.,
        'h': 1. / 2,
        'q': 1. / 4,
        'e': 1. / 8,
        's': 1. / 16,
    }



from mindpile.Mapping.types import List
from mindpile.Mapping.utils import MethodCall

@MethodCall(target="X3.Lib:GlobalConstBoolean", valueIn=bool, valueOut=bool)
@MethodCall(target="X3.Lib:GlobalConstBooleanArray", valueIn=List(bool), valueOut=List(bool))
@MethodCall(target="X3.Lib:GlobalConstSingle", valueIn=float, valueOut=float)
@MethodCall(target="X3.Lib:GlobalConstNumericArray", valueIn=List(float), valueOut=List(float))
@MethodCall(target="X3.Lib:GlobalConstString", valueIn=str, valueOut=str)
def const():
    return '''
        valueOut = valueIn
    '''



import logging
from ev3dev2.motor import MoveTank
from ev3dev2.sensor.lego import InfraredSensor
from time import sleep

log = logging.getLogger(__name__)


# ============
# Tank classes
# ============
class RemoteControlledTank(MoveTank):
    def __init__(self, left_motor_port, right_motor_port, polarity='inversed', speed=400, channel=1):
        MoveTank.__init__(self, left_motor_port, right_motor_port)
        self.set_polarity(polarity)

        left_motor = self.motors[left_motor_port]
        right_motor = self.motors[right_motor_port]
        self.speed_sp = speed
        self.remote = InfraredSensor()
        self.remote.on_channel1_top_left = self.make_move(left_motor, self.speed_sp)
        self.remote.on_channel1_bottom_left = self.make_move(left_motor, self.speed_sp * -1)
        self.remote.on_channel1_top_right = self.make_move(right_motor, self.speed_sp)
        self.remote.on_channel1_bottom_right = self.make_move(right_motor, self.speed_sp * -1)
        self.channel = channel

    def make_move(self, motor, dc_sp):
        def move(state):
            if state:
                motor.run_forever(speed_sp=dc_sp)
            else:
                motor.stop()

        return move

    def main(self):

        try:
            while True:
                self.remote.process()
                sleep(0.01)

        # Exit cleanly so that all motors are stopped
        except (KeyboardInterrupt, Exception) as e:
            log.exception(e)
            self.off()



NAMESPACE = "{http://www.ni.com/VirtualInstrument.xsd}"


def addNameSpace(string):
    return NAMESPACE + string


def removeNameSpace(string):
    return "".join(string.split("}")[1::])


def getDepthInfoTree(elem, depth=0):
    return (depth, elem, [getDepthInfoTree(child, depth=depth + 1) for child in elem])


def addSpacing(spaces, text):
    return "\n".join([(" " * spaces) + text for text in text.split("\n")])

